from unittest import TestCase
from itertools import product
from collections import defaultdict

from dark.aa import (
    AminoAcid,
    find,
    propertiesForSequence,
    clustersForSequence,
    compareAaReads,
    matchToString,
)

from dark.aaVars import (
    PROPERTIES,
    ALL_PROPERTIES,
    PROPERTY_NAMES,
    ACIDIC,
    ALIPHATIC,
    AROMATIC,
    BASIC_POSITIVE,
    HYDROPHILIC,
    HYDROPHOBIC,
    HYDROXYLIC,
    NEGATIVE,
    NONE,
    POLAR,
    SMALL,
    SULPHUR,
    TINY,
    NAMES,
    NAMES_TO_ABBREV1,
    ABBREV3,
    ABBREV3_TO_ABBREV1,
    AA_LETTERS,
    PROPERTY_CLUSTERS,
    PROPERTY_DETAILS_RAW,
    CODONS,
    REVERSE_CODONS,
    START_CODON,
    STOP_CODONS,
)
from dark.reads import AARead


class TestAALetters(TestCase):
    """
    Test the AA_LETTERS string in dark.aa
    """

    def testExpectedAALetters(self):
        """
        The AA_LETTERS value must be as expected.
        """
        # From https://en.wikipedia.org/wiki/Amino_acid
        expected = sorted("ARNDCEQGHILKMFPSTWYV")
        self.assertEqual(expected, AA_LETTERS)


class TestProperties(TestCase):
    """
    Test the AA properties in dark.aa
    """

    def testCorrectAAs(self):
        """
        The PROPERTIES dict must have the correct AA keys.
        """
        self.assertEqual(AA_LETTERS, sorted(PROPERTIES))

    def testAllProperties(self):
        """
        The ALL_PROPERTIES tuple must contain all known properties.
        """
        expected = set(
            [
                ACIDIC,
                ALIPHATIC,
                AROMATIC,
                BASIC_POSITIVE,
                HYDROPHILIC,
                HYDROPHOBIC,
                HYDROXYLIC,
                NEGATIVE,
                NONE,
                POLAR,
                SMALL,
                SULPHUR,
                TINY,
            ]
        )
        self.assertEqual(set(ALL_PROPERTIES), expected)

    def testPropertyValuesDiffer(self):
        """
        All individual property values must be different.
        """
        self.assertEqual(len(ALL_PROPERTIES), len(set(ALL_PROPERTIES)))

    def testEachAAHasItsCorrectPropertiesAndNoOthers(self):
        """
        Each amino acid must have the properties expected of it, and no others.
        """
        expected = {
            "A": set([HYDROPHOBIC, SMALL, TINY]),
            "C": set([HYDROPHOBIC, SMALL, TINY, SULPHUR]),
            "D": set([HYDROPHILIC, SMALL, POLAR, NEGATIVE]),
            "E": set([HYDROPHILIC, NEGATIVE, ACIDIC]),
            "F": set([HYDROPHOBIC, AROMATIC]),
            "G": set([HYDROPHILIC, SMALL, TINY]),
            "H": set([HYDROPHOBIC, AROMATIC, POLAR, BASIC_POSITIVE]),
            "I": set([ALIPHATIC, HYDROPHOBIC]),
            "K": set([HYDROPHOBIC, BASIC_POSITIVE, POLAR]),
            "L": set([ALIPHATIC, HYDROPHOBIC]),
            "M": set([HYDROPHOBIC, SULPHUR]),
            "N": set([HYDROPHILIC, SMALL, POLAR, ACIDIC]),
            "P": set([HYDROPHILIC, SMALL]),
            "Q": set([HYDROPHILIC, POLAR, ACIDIC]),
            "R": set([HYDROPHILIC, POLAR, BASIC_POSITIVE]),
            "S": set([HYDROPHILIC, SMALL, POLAR, HYDROXYLIC]),
            "T": set([HYDROPHOBIC, SMALL, HYDROXYLIC]),
            "V": set([ALIPHATIC, HYDROPHOBIC, SMALL]),
            "W": set([HYDROPHOBIC, AROMATIC, POLAR]),
            "Y": set([HYDROPHOBIC, AROMATIC, POLAR]),
        }

        # Make sure our 'expected' dict (above) has properties for everything.
        self.assertEqual(AA_LETTERS, sorted(expected))

        for aa in AA_LETTERS:
            for prop in ALL_PROPERTIES:
                if prop in expected[aa]:
                    self.assertTrue(bool(PROPERTIES[aa] & prop))
                else:
                    self.assertFalse(bool(PROPERTIES[aa] & prop))

    def testProperyNamesKeys(self):
        """
        The PROPERTY_NAMES dict must have a key for each property.
        """
        self.assertEqual(sorted(ALL_PROPERTIES), sorted(PROPERTY_NAMES))

    def testProperyNamesValuesDiffer(self):
        """
        The PROPERTY_NAMES dict must have different values for each key.
        """
        self.assertEqual(len(PROPERTY_NAMES), len(set(PROPERTY_NAMES.values())))


class TestNames(TestCase):
    """
    Test the NAMES dict in dark.aa
    """

    def testCorrectAAs(self):
        """
        The NAMES dict must have the correct AA keys.
        """
        self.assertEqual(AA_LETTERS, sorted(NAMES))

    def testValuesDiffer(self):
        """
        All individual names must be different.
        """
        names = list(NAMES.values())
        self.assertEqual(len(names), len(set(names)))


class TestNamesToAbbrev1(TestCase):
    """
    Test the NAMES_TO_ABBREV1 dict in dark.aa
    """

    def testCorrectAAs(self):
        """
        The NAMES_TO_ABBREV1 dict must have the correct AA values.
        """
        self.assertEqual(AA_LETTERS, sorted(NAMES_TO_ABBREV1.values()))


class TestAbbrev3(TestCase):
    """
    Test the ABBREV3 dict in dark.aa
    """

    def testCorrectAAs(self):
        """
        The ABBREV3 dict must have the correct AA keys.
        """
        self.assertEqual(AA_LETTERS, sorted(ABBREV3))

    def testValuesDiffer(self):
        """
        All individual names must be different.
        """
        names = list(ABBREV3.values())
        self.assertEqual(len(names), len(set(names)))


class TestAbbrev3ToAbbrev1(TestCase):
    """
    Test the ABBREV3_TO_ABBREV1 dict in dark.aa
    """

    def testCorrectAAs(self):
        """
        The ABBREV3_TO_ABBREV1 dict must have the correct AA values.
        """
        self.assertEqual(AA_LETTERS, sorted(ABBREV3_TO_ABBREV1.values()))


class TestCodons(TestCase):
    """
    Tests for the CODONS dict.
    """

    def testCorrectAAs(self):
        """
        The CODONS dict must have the correct AA keys.
        """
        self.assertEqual(AA_LETTERS, sorted(CODONS))

    def testNumberCodons(self):
        """
        The table must contain 61 codons. This is 4^3 - 3. I.e., all possible
        combinations of 3 bases minus the three stop codons.
        """
        self.assertEqual(61, sum(len(codons) for codons in CODONS.values()))

    def testCodonLength(self):
        """
        All codons must be three bases long.
        """
        for codons in CODONS.values():
            self.assertTrue(all(len(codon) == 3 for codon in codons))

    def testCodonContent(self):
        """
        Codons must only contain the letters A, T, G, C.
        """
        for codons in CODONS.values():
            for codon in codons:
                self.assertTrue(all(letter in "ACGT" for letter in codon))

    def testCodonsAreNotAbbrev3s(self):
        """
        No codon can be the same as an amino acid 3-letter abbreviation (or
        else our find function may not be unambiguous in what it returns).
        """
        for codons in CODONS.values():
            self.assertFalse(
                any(codon.title() in ABBREV3_TO_ABBREV1 for codon in codons)
            )

    def testDistinct(self):
        """
        All nucleotide triples must be distinct.
        """
        seen = set()
        for codons in CODONS.values():
            for codon in codons:
                seen.add(codon)
        self.assertEqual(61, len(seen))

    def testStartCodon(self):
        """
        The start codon must have the expected value.
        """
        self.assertEqual("ATG", START_CODON)

    def testStartCodonIsMethionine(self):
        """
        The start codon is the same as Methionine.
        """
        self.assertEqual((START_CODON,), CODONS["M"])

    def testStopCodons(self):
        """
        The stop codons must have the expected value.
        """
        self.assertEqual(("TAA", "TAG", "TGA"), STOP_CODONS)

    def testStopCodonsNotInCodonTable(self):
        """
        The stop codons must not be in the main table.
        """
        for stop in STOP_CODONS:
            for codons in CODONS.values():
                self.assertNotIn(stop, codons)

    def testAllCodonsPresent(self):
        """
        All possible codons must be present, as either coding for an AA
        or as a stop codon.
        """
        combinations = set("".join(x) for x in product("ACGT", "ACGT", "ACGT"))
        for codons in CODONS.values():
            for codon in codons:
                combinations.remove(codon)

        # Just the stop codons should be left.
        self.assertEqual(set(STOP_CODONS), combinations)

    def testReverseCodons(self):
        """
        The REVERSE_CODONS dictionary must be correct.
        """
        roundtrip = defaultdict(set)
        for codon, aa in REVERSE_CODONS.items():
            roundtrip[aa].add(codon)

        for aa, codons in CODONS.items():
            self.assertEqual(roundtrip[aa], set(codons), aa)


class TestAminoAcid(TestCase):
    """
    Test the AminoAcid class in dark.aa
    """

    def testExpectedAttributes(self):
        """
        An amino acid instance must have the expected attributes.
        """
        properties = {}
        propertyDetails = {}
        codons = []
        propertyClusters = {}
        aa = AminoAcid(
            "Alanine", "Ala", "A", codons, properties, propertyDetails, propertyClusters
        )
        self.assertEqual("Alanine", aa.name)
        self.assertEqual("Ala", aa.abbrev3)
        self.assertEqual("A", aa.abbrev1)
        self.assertIs(codons, aa.codons)
        self.assertIs(properties, aa.properties)
        self.assertIs(propertyDetails, aa.propertyDetails)
        self.assertIs(propertyClusters, aa.propertyClusters)


class TestFind(TestCase):
    """
    Test the find function in dark.aa
    """

    def testFindUnknown(self):
        """
        find must return an empty generator when called with an unrecognized
        value.
        """
        self.assertEqual([], list(find("silly")))

    def testFindByAbbrev1(self):
        """
        It must be possible to find an amino acid by its 1-letter abbreviation.
        """
        aa = list(find("A"))[0]
        self.assertEqual("Alanine", aa.name)
        self.assertEqual("Ala", aa.abbrev3)
        self.assertEqual("A", aa.abbrev1)
        self.assertEqual(("GCA", "GCC", "GCG", "GCT"), aa.codons)
        self.assertEqual(HYDROPHOBIC | SMALL | TINY, aa.properties)
        self.assertEqual(
            {
                "aliphaticity": 0.305785123967,
                "aromaticity": -0.550128534704,
                "composition": -1.0,
                "hydrogenation": 0.8973042362,
                "hydropathy": 0.4,
                "hydroxythiolation": -0.265160523187,
                "iep": -0.191489361702,
                "polar requirement": -0.463414634146,
                "polarity": -0.20987654321,
                "volume": -0.664670658683,
            },
            aa.propertyDetails,
        )
        self.assertEqual(
            {
                "aliphaticity": 1,
                "aromaticity": 1,
                "composition": 1,
                "hydrogenation": 1,
                "hydropathy": 3,
                "hydroxythiolation": 2,
                "iep": 2,
                "polar requirement": 2,
                "polarity": 2,
                "volume": 2,
            },
            aa.propertyClusters,
        )

    def testFindByAbbrev3(self):
        """
        It must be possible to find an amino acid by its 3-letter abbreviation.
        """
        aa = list(find("Ala"))[0]
        self.assertEqual("Alanine", aa.name)
        self.assertEqual("Ala", aa.abbrev3)
        self.assertEqual("A", aa.abbrev1)
        self.assertEqual(("GCA", "GCC", "GCG", "GCT"), aa.codons)
        self.assertEqual(HYDROPHOBIC | SMALL | TINY, aa.properties)
        self.assertEqual(
            {
                "aliphaticity": 0.305785123967,
                "aromaticity": -0.550128534704,
                "composition": -1.0,
                "hydrogenation": 0.8973042362,
                "hydropathy": 0.4,
                "hydroxythiolation": -0.265160523187,
                "iep": -0.191489361702,
                "polar requirement": -0.463414634146,
                "polarity": -0.20987654321,
                "volume": -0.664670658683,
            },
            aa.propertyDetails,
        )
        self.assertEqual(
            {
                "aliphaticity": 1,
                "aromaticity": 1,
                "composition": 1,
                "hydrogenation": 1,
                "hydropathy": 3,
                "hydroxythiolation": 2,
                "iep": 2,
                "polar requirement": 2,
                "polarity": 2,
                "volume": 2,
            },
            aa.propertyClusters,
        )

    def testFindByCodon(self):
        """
        It must be possible to find an amino acid by a 3-letter codon.
        """
        aa = list(find("GCC"))[0]
        self.assertEqual("Alanine", aa.name)
        self.assertEqual("Ala", aa.abbrev3)
        self.assertEqual("A", aa.abbrev1)
        self.assertEqual(("GCA", "GCC", "GCG", "GCT"), aa.codons)
        self.assertEqual(HYDROPHOBIC | SMALL | TINY, aa.properties)
        self.assertEqual(
            {
                "aliphaticity": 0.305785123967,
                "aromaticity": -0.550128534704,
                "composition": -1.0,
                "hydrogenation": 0.8973042362,
                "hydropathy": 0.4,
                "hydroxythiolation": -0.265160523187,
                "iep": -0.191489361702,
                "polar requirement": -0.463414634146,
                "polarity": -0.20987654321,
                "volume": -0.664670658683,
            },
            aa.propertyDetails,
        )
        self.assertEqual(
            {
                "aliphaticity": 1,
                "aromaticity": 1,
                "composition": 1,
                "hydrogenation": 1,
                "hydropathy": 3,
                "hydroxythiolation": 2,
                "iep": 2,
                "polar requirement": 2,
                "polarity": 2,
                "volume": 2,
            },
            aa.propertyClusters,
        )

    def testFindByName(self):
        """
        It must be possible to find an amino acid by its name.
        """
        aa = list(find("Alanine"))[0]
        self.assertEqual("Alanine", aa.name)
        self.assertEqual("Ala", aa.abbrev3)
        self.assertEqual("A", aa.abbrev1)
        self.assertEqual(("GCA", "GCC", "GCG", "GCT"), aa.codons)
        self.assertEqual(HYDROPHOBIC | SMALL | TINY, aa.properties)
        self.assertEqual(
            {
                "aliphaticity": 0.305785123967,
                "aromaticity": -0.550128534704,
                "composition": -1.0,
                "hydrogenation": 0.8973042362,
                "hydropathy": 0.4,
                "hydroxythiolation": -0.265160523187,
                "iep": -0.191489361702,
                "polar requirement": -0.463414634146,
                "polarity": -0.20987654321,
                "volume": -0.664670658683,
            },
            aa.propertyDetails,
        )
        self.assertEqual(
            {
                "aliphaticity": 1,
                "aromaticity": 1,
                "composition": 1,
                "hydrogenation": 1,
                "hydropathy": 3,
                "hydroxythiolation": 2,
                "iep": 2,
                "polar requirement": 2,
                "polarity": 2,
                "volume": 2,
            },
            aa.propertyClusters,
        )

    def testFindByNameCaseIgnored(self):
        """
        It must be possible to find an amino acid by its name when the name is
        given in mixed case.
        """
        aa = list(find("alaNIne"))[0]
        self.assertEqual("Alanine", aa.name)
        self.assertEqual("Ala", aa.abbrev3)
        self.assertEqual("A", aa.abbrev1)
        self.assertEqual(("GCA", "GCC", "GCG", "GCT"), aa.codons)
        self.assertEqual(HYDROPHOBIC | SMALL | TINY, aa.properties)
        self.assertEqual(
            {
                "aliphaticity": 0.305785123967,
                "aromaticity": -0.550128534704,
                "composition": -1.0,
                "hydrogenation": 0.8973042362,
                "hydropathy": 0.4,
                "hydroxythiolation": -0.265160523187,
                "iep": -0.191489361702,
                "polar requirement": -0.463414634146,
                "polarity": -0.20987654321,
                "volume": -0.664670658683,
            },
            aa.propertyDetails,
        )
        self.assertEqual(
            {
                "aliphaticity": 1,
                "aromaticity": 1,
                "composition": 1,
                "hydrogenation": 1,
                "hydropathy": 3,
                "hydroxythiolation": 2,
                "iep": 2,
                "polar requirement": 2,
                "polarity": 2,
                "volume": 2,
            },
            aa.propertyClusters,
        )

    def testFindByNameCaseIgnoredNameWithSpace(self):
        """
        It must be possible to find an amino acid by its name when the name is
        given in mixed case, including if the name has a space in it.
        """
        aa = list(find("asparTIC aCId"))[0]
        self.assertEqual("Aspartic acid", aa.name)
        self.assertEqual("Asp", aa.abbrev3)
        self.assertEqual("D", aa.abbrev1)
        self.assertEqual(("GAC", "GAT"), aa.codons)
        self.assertEqual(HYDROPHILIC | SMALL | POLAR | NEGATIVE, aa.properties)
        self.assertEqual(
            {
                "aliphaticity": -0.818181818182,
                "aromaticity": -1.0,
                "composition": 0.00363636363636,
                "hydrogenation": -0.90243902439,
                "hydropathy": -0.777777777778,
                "hydroxythiolation": -0.348394768133,
                "iep": -1.0,
                "polar requirement": 1.0,
                "polarity": 1.0,
                "volume": -0.389221556886,
            },
            aa.propertyDetails,
        )
        self.assertEqual(
            {
                "aliphaticity": 1,
                "aromaticity": 1,
                "composition": 2,
                "hydrogenation": 1,
                "hydropathy": 1,
                "hydroxythiolation": 2,
                "iep": 1,
                "polar requirement": 4,
                "polarity": 4,
                "volume": 3,
            },
            aa.propertyClusters,
        )

    def testFindByPartialName(self):
        """
        It must be possible to find an amino acid by a partial name.
        """
        aa = list(find("nine"))[0]
        self.assertEqual("Alanine", aa.name)

    def testFindByPartialNameMixedCase(self):
        """
        It must be possible to find an amino acid by a mixed-case partial name.
        """
        aa = list(find("NiNe"))[0]
        self.assertEqual("Alanine", aa.name)

    def testFindMultipleMatches(self):
        """
        It must be possible to find more than one matching amino acid.
        """
        aa1, aa2 = list(find("acid"))
        self.assertEqual("Aspartic acid", aa1.name)
        self.assertEqual("Glutamic acid", aa2.name)


class TestPropertyClusters(TestCase):
    """
    Tests for the PROPERTY_CLUSTERS dict.
    """

    def testPropertyClusterKeys(self):
        """
        The PROPERTY_CLUSTERS dict must contain the right keys.
        """
        self.assertEqual(AA_LETTERS, sorted(PROPERTY_CLUSTERS))

    def testNumberOfValues(self):
        """
        Each key in PROPERTY_CLUSTERS must have a dict with 10 elements in it
        as the value.
        """
        for propertyNames in PROPERTY_CLUSTERS.values():
            self.assertEqual(
                [
                    "aliphaticity",
                    "aromaticity",
                    "composition",
                    "hydrogenation",
                    "hydropathy",
                    "hydroxythiolation",
                    "iep",
                    "polar requirement",
                    "polarity",
                    "volume",
                ],
                sorted(propertyNames),
            )

    def testAliphaticity(self):
        """
        Aliphaticity must always be in cluster 1.
        """
        for propertiesDict in PROPERTY_CLUSTERS.values():
            self.assertEqual(1, propertiesDict["aliphaticity"])

    def testPermittedClustersOnly(self):
        """
        All cluster numbers must be in {1, 2, 3, 4, 5}.
        """
        for propertiesDict in PROPERTY_CLUSTERS.values():
            for clusterNr in propertiesDict.values():
                self.assertIn(clusterNr, {1, 2, 3, 4, 5})


class TestPropertyDetailsRaw(TestCase):
    """
    Tests for the PROPERTY_DETAILS_RAW dict.
    """

    def testPropertyDetailsRawKeys(self):
        """
        The PROPERTY_DETAILS_RAW dict must contain the right keys.
        """
        self.assertEqual(AA_LETTERS, sorted(PROPERTY_DETAILS_RAW))

    def testNumberOfValues(self):
        """
        Each key in PROPERTY_DETAILS_RAW must have a dict with 10 elements in
        it as the value.
        """
        for propertyNames in PROPERTY_DETAILS_RAW.values():
            self.assertEqual(
                [
                    "aliphaticity",
                    "aromaticity",
                    "composition",
                    "hydrogenation",
                    "hydropathy",
                    "hydroxythiolation",
                    "iep",
                    "polar requirement",
                    "polarity",
                    "volume",
                ],
                sorted(propertyNames),
            )

    def testAliphaticity(self):
        """
        Aliphaticity of Alanin must have the right value.
        """
        self.assertEqual(0.239, PROPERTY_DETAILS_RAW["A"]["aliphaticity"])

    def testValuesMustBeFloats(self):
        """
        Each key in PROPERTY_DETAILS_RAW must have a dict whose values are
        all floats.
        """
        for properties in PROPERTY_DETAILS_RAW.values():
            self.assertTrue(all(isinstance(v, float) for v in properties.values()))


class TestPropertiesForSequence(TestCase):
    """
    Tests for the propertiesForSequence function in aa.py
    """

    def testUnknownProperty(self):
        """
        A C{ValueError} must be raised if an unknown property name is passed.
        """
        error = "Unknown property: xxx"
        read = AARead("id", "RRR")
        self.assertRaisesRegex(ValueError, error, propertiesForSequence, read, ["xxx"])

    def testNoProperties(self):
        """
        If no properties are wanted, an empty dict must be returned.
        """
        read = AARead("id", "RRR")
        self.assertEqual({}, propertiesForSequence(read, []))

    def testOnePropertyEmptySequence(self):
        """
        If one property is wanted but the sequence is empty, a dict with the
        property must be returned, and have an empty list value.
        """
        read = AARead("id", "")
        self.assertEqual(
            {
                "hydropathy": [],
            },
            propertiesForSequence(read, ["hydropathy"]),
        )

    def testPropertyNameIsLowercased(self):
        """
        The property name must be lower-cased in the result.
        """
        read = AARead("id", "")
        self.assertTrue("hydropathy" in propertiesForSequence(read, ["HYDROPATHY"]))

    def testOneProperty(self):
        """
        If one property is wanted, a dict with the property must be returned,
        and have the expected property values.
        """
        read = AARead("id", "AI")
        self.assertEqual(
            {
                "hydropathy": [0.4, 1.0],
            },
            propertiesForSequence(read, ["hydropathy"]),
        )

    def testTwoProperties(self):
        """
        If two properties are wanted, a dict with the properties must be
        returned, and have the expected property values.
        """
        read = AARead("id", "AI")
        self.assertEqual(
            {
                "composition": [-1.0, -1.0],
                "hydropathy": [0.4, 1.0],
            },
            propertiesForSequence(read, ["composition", "hydropathy"]),
        )

    def testDuplicatedPropertyName(self):
        """
        If a property name is mentioned more than once, a dict with the
        wanted properties must be returned, and have the expected property
        values.
        """
        read = AARead("id", "AI")
        self.assertEqual(
            {
                "composition": [-1.0, -1.0],
                "hydropathy": [0.4, 1.0],
            },
            propertiesForSequence(read, ["composition", "hydropathy", "hydropathy"]),
        )

    def testMissingAminoAcid(self):
        """
        If an unknown amino acid appears in the sequence, its property value
        must be the default (-1.1).
        """
        read = AARead("id", "XX")
        self.assertEqual(
            {
                "composition": [-1.1, -1.1],
                "hydropathy": [-1.1, -1.1],
            },
            propertiesForSequence(read, ["composition", "hydropathy"]),
        )

    def testMissingAminoAcidWithNonDefaultMissingValue(self):
        """
        If an unknown amino acid appears in the sequence, its property value
        must be the missing AA value that was passed.
        """
        read = AARead("id", "XX")
        self.assertEqual(
            {
                "composition": [-1.5, -1.5],
                "hydropathy": [-1.5, -1.5],
            },
            propertiesForSequence(
                read, ["composition", "hydropathy"], missingAAValue=-1.5
            ),
        )


class TestClustersForSequence(TestCase):
    """
    Tests for the clustersForSequence function in aa.py
    """

    def testUnknownProperty(self):
        """
        A C{ValueError} must be raised if an unknown property name is passed.
        """
        error = "Unknown property: xxx"
        read = AARead("id", "RRR")
        self.assertRaisesRegex(ValueError, error, clustersForSequence, read, ["xxx"])

    def testNoProperties(self):
        """
        If no properties are wanted, an empty dict must be returned.
        """
        read = AARead("id", "RRR")
        self.assertEqual({}, clustersForSequence(read, []))

    def testOnePropertyEmptySequence(self):
        """
        If one property is wanted but the sequence is empty, a dict with the
        property must be returned, and have an empty list value.
        """
        read = AARead("id", "")
        self.assertEqual(
            {
                "hydropathy": [],
            },
            clustersForSequence(read, ["hydropathy"]),
        )

    def testPropertyNameIsLowercased(self):
        """
        The property name must be lower-cased in the result.
        """
        read = AARead("id", "")
        self.assertTrue("hydropathy" in clustersForSequence(read, ["HYDROPATHY"]))

    def testOneProperty(self):
        """
        If one property is wanted, a dict with the property must be returned,
        and have the expected property cluster number.
        """
        read = AARead("id", "AI")
        self.assertEqual(
            {
                "hydropathy": [3, 4],
            },
            clustersForSequence(read, ["hydropathy"]),
        )

    def testTwoProperties(self):
        """
        If two properties are wanted, a dict with the properties must be
        returned, and have the expected property cluster numbers.
        """
        read = AARead("id", "AI")
        self.assertEqual(
            {
                "composition": [1, 1],
                "hydropathy": [3, 4],
            },
            clustersForSequence(read, ["composition", "hydropathy"]),
        )

    def testDuplicatedPropertyName(self):
        """
        If a property name is mentioned more than once, a dict with the
        wanted properties must be returned, and have the expected property
        cluster numbers.
        """
        read = AARead("id", "AI")
        self.assertEqual(
            {
                "composition": [1, 1],
                "hydropathy": [3, 4],
            },
            clustersForSequence(read, ["composition", "hydropathy", "hydropathy"]),
        )

    def testMissingAminoAcid(self):
        """
        If an unknown amino acid appears in the sequence, its property cluster
        must be the default (0).
        """
        read = AARead("id", "XX")
        self.assertEqual(
            {
                "composition": [0, 0],
                "hydropathy": [0, 0],
            },
            clustersForSequence(read, ["composition", "hydropathy"]),
        )

    def testMissingAminoAcidWithNonDefaultMissingValue(self):
        """
        If an unknown amino acid appears in the sequence, its property cluster
        must be the missing AA value that was passed.
        """
        read = AARead("id", "XX")
        self.assertEqual(
            {
                "composition": [10, 10],
                "hydropathy": [10, 10],
            },
            clustersForSequence(read, ["composition", "hydropathy"], missingAAValue=10),
        )


class TestCompareAaReads(TestCase):
    """
    Test the compareAaReads function.
    """

    def testEmptySequences(self):
        """
        Two empty sequences must compare as expected.
        """
        self.assertEqual(
            {
                "match": {
                    "matchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                },
                "read1": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
                "read2": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
            },
            compareAaReads(AARead("id1", ""), AARead("id2", "")),
        )

    def testExactMatch(self):
        """
        Two sequences that match exactly must compare as expected.
        """
        self.assertEqual(
            {
                "match": {
                    "matchCount": 5,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                },
                "read1": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
                "read2": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
            },
            compareAaReads(AARead("id1", "GALHN"), AARead("id2", "GALHN")),
        )

    def testMismatch(self):
        """
        If the sequences have mismatched bases, their count
        must be given correctly in the nonGapMismatchCount.
        """
        self.assertEqual(
            {
                "match": {
                    "matchCount": 3,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 2,
                },
                "read1": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
                "read2": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
            },
            compareAaReads(AARead("id1", "GALYY"), AARead("id2", "GALHN")),
        )

    def testNoOffsets(self):
        """
        If an empty set of wanted offsets is passed, the result must be empty.
        """
        self.assertEqual(
            {
                "match": {
                    "matchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                },
                "read1": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
                "read2": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
            },
            compareAaReads(
                AARead("id1", "GAL-N"), AARead("id2", "G-LHN"), offsets=set()
            ),
        )

    def testOffsets(self):
        """
        If a set of wanted offsets is passed, the result must be restricted to
        just those offsets.
        """
        self.assertEqual(
            {
                "match": {
                    "matchCount": 1,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 1,
                },
                "read1": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
                "read2": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
            },
            compareAaReads(
                AARead("id1", "GAL-L"), AARead("id2", "G-LHN"), offsets=set([0, 4])
            ),
        )

    def testGapInFirst(self):
        """
        A gap in the first sequence must be dealt with correctly.
        """
        self.assertEqual(
            {
                "match": {
                    "matchCount": 4,
                    "gapMismatchCount": 1,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                },
                "read1": {
                    "extraCount": 0,
                    "gapOffsets": [3],
                },
                "read2": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
            },
            compareAaReads(AARead("id1", "GAL-N"), AARead("id2", "GALHN")),
        )

    def testGapInSecond(self):
        """
        A gap in the second sequence must be dealt with correctly.
        """
        self.assertEqual(
            {
                "match": {
                    "matchCount": 3,
                    "gapMismatchCount": 2,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                },
                "read1": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
                "read2": {
                    "extraCount": 0,
                    "gapOffsets": [1, 2],
                },
            },
            compareAaReads(AARead("id1", "GALHN"), AARead("id2", "G--HN")),
        )

    def testNonDefaultGapChars(self):
        """
        We must be able to specify the gap characters.
        """
        for gap in "+$":
            self.assertEqual(
                {
                    "match": {
                        "matchCount": 3,
                        "gapMismatchCount": 2,
                        "gapGapMismatchCount": 0,
                        "nonGapMismatchCount": 0,
                    },
                    "read1": {
                        "extraCount": 0,
                        "gapOffsets": [2],
                    },
                    "read2": {
                        "extraCount": 0,
                        "gapOffsets": [0],
                    },
                },
                compareAaReads(
                    AARead("id1", "GA%sHN" % gap),
                    AARead("id2", "%sALHN" % gap),
                    gapChars="+$",
                ),
            )

    def testGapGap(self):
        """
        Coinciding gaps in the sequences must be dealt with correctly
        """
        self.assertEqual(
            {
                "match": {
                    "matchCount": 2,
                    "gapMismatchCount": 2,
                    "gapGapMismatchCount": 1,
                    "nonGapMismatchCount": 0,
                },
                "read1": {
                    "extraCount": 0,
                    "gapOffsets": [2, 3],
                },
                "read2": {
                    "extraCount": 0,
                    "gapOffsets": [1, 2],
                },
            },
            compareAaReads(AARead("id1", "GA--N"), AARead("id2", "G--HN")),
        )

    def testExtraInFirst(self):
        """
        If the first sequence has extra bases, they must be indicated in the
        extraCount.
        """
        self.assertEqual(
            {
                "match": {
                    "matchCount": 5,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                },
                "read1": {
                    "extraCount": 2,
                    "gapOffsets": [],
                },
                "read2": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
            },
            compareAaReads(AARead("id1", "GALHNHN"), AARead("id2", "GALHN")),
        )

    def testExtraInSecond(self):
        """
        If the second sequence has extra bases, they must be indicated in the
        extraCount.
        """
        self.assertEqual(
            {
                "match": {
                    "matchCount": 5,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                },
                "read1": {
                    "extraCount": 0,
                    "gapOffsets": [],
                },
                "read2": {
                    "extraCount": 2,
                    "gapOffsets": [],
                },
            },
            compareAaReads(AARead("id1", "GALHN"), AARead("id2", "GALHNHN")),
        )


class TestMatchToString(TestCase):
    """
    Test the matchToString function.
    """

    def testMismatchAndMatch(self):
        """
        Two sequences containing matches and mismatches must compare as
        expected.
        """
        read1 = AARead("id1", "GALHNG")
        read2 = AARead("id2", "GALHNA")
        match = compareAaReads(read1, read2)

        self.assertEqual(
            """\
Matches:
  Overall: 5/6 (83.33%)
  Not involving gaps (i.e., identities): 5/6 (83.33%)
Mismatches:
  Overall: 1/6 (16.67%)
  Not involving gaps (i.e., conflicts): 1/6 (16.67%)
  Involving a gap in one sequence: 0
  Involving a gap in both sequences: 0
Id: id1
  Length: 6
  Gaps: 0
Id: id2
  Length: 6
  Gaps: 0""",
            matchToString(match, read1, read2),
        )

    def testGapAndMatch(self):
        """
        Two sequences containing gaps and matches must compare as
        expected.
        """
        read1 = AARead("id1", "GALHN-")
        read2 = AARead("id2", "GALHNA")
        match = compareAaReads(read1, read2)

        self.assertEqual(
            """\
Matches:
  Overall: 5/6 (83.33%)
  Not involving gaps (i.e., identities): 5/5 (100.00%)
Mismatches:
  Overall: 1/6 (16.67%)
  Not involving gaps (i.e., conflicts): 0
  Involving a gap in one sequence: 1/6 (16.67%)
  Involving a gap in both sequences: 0
Id: id1
  Length: 6
  Gaps: 1/6 (16.67%)
  Gap locations (1-based): 6
Id: id2
  Length: 6
  Gaps: 0""",
            matchToString(match, read1, read2),
        )

    def testGapAndMatchAndMismatch(self):
        """
        Two sequences containing gaps and matches and mismatches must compare
        as expected.
        """
        read1 = AARead("id1", "--GBLHN-")
        read2 = AARead("id2", "--GALHNA")
        match = compareAaReads(read1, read2)

        self.assertEqual(
            """\
Matches:
  Overall: 4/8 (50.00%)
  Not involving gaps (i.e., identities): 4/5 (80.00%)
Mismatches:
  Overall: 4/8 (50.00%)
  Not involving gaps (i.e., conflicts): 1/8 (12.50%)
  Involving a gap in one sequence: 1/8 (12.50%)
  Involving a gap in both sequences: 2/8 (25.00%)
Id: id1
  Length: 8
  Gaps: 3/8 (37.50%)
  Gap locations (1-based): 1, 2, 8
Id: id2
  Length: 8
  Gaps: 2/8 (25.00%)
  Gap locations (1-based): 1, 2""",
            matchToString(match, read1, read2),
        )

    def testUnequalLengthGapAndMatchAndMismatch(self):
        """
        Two sequences of unequal length containing gaps and matches and
        mismatches must compare as expected.
        """
        read1 = AARead("id1", "--GBLHN-CC")
        read2 = AARead("id2", "--GALHNA")
        match = compareAaReads(read1, read2)

        self.assertEqual(
            """\
Matches:
  Overall: 4/10 (40.00%) of sequence 1, 4/8 (50.00%) of sequence 2
  Not involving gaps (i.e., identities): 4/5 (80.00%)
Mismatches:
  Overall: 4/10 (40.00%) of sequence 1, 4/8 (50.00%) of sequence 2
  Not involving gaps (i.e., conflicts): 1/10 (10.00%) of sequence 1, \
1/8 (12.50%) of sequence 2
  Involving a gap in one sequence: 1/10 (10.00%) of sequence 1, \
1/8 (12.50%) of sequence 2
  Involving a gap in both sequences: 2/10 (20.00%) of sequence 1, \
2/8 (25.00%) of sequence 2
Id: id1
  Length: 10
  Gaps: 3/10 (30.00%)
  Gap locations (1-based): 1, 2, 8
  Extra amino acids at end: 2/10 (20.00%)
Id: id2
  Length: 8
  Gaps: 2/8 (25.00%)
  Gap locations (1-based): 1, 2""",
            matchToString(match, read1, read2),
        )



#!/usr/bin/env python

import argparse

from dark.civ.proteins import SqliteIndex


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description="Print a description of a protein/genome database.",
)

parser.add_argument("databaseFile", help="The Sqlite3 database file.")

args = parser.parse_args()

db = SqliteIndex(args.databaseFile)

print("%d genomes with %d proteins" % (db.genomeCount(), db.proteinCount()))



from operator import itemgetter


class ColorsForCounts:
    """
    Maintain a collection of count thresholds and colors with methods to get a
    color or a CSS name for a count.

    @param colors: An C{iterable} of space separated "value color" strings,
        such as ["100 red", "200 rgb(23, 190, 207)", "700 #CF3CF3"]. Or C{None}
        if no colors (other than C{defaultColor}) should be used.
    @param defaultColor: The C{str} color to use for counts that do not reach
        the lowest count threshold for any color in C{colors}.
    @raise ValueError: If an incorrect count/color pair is found in C{colors}.
    """

    def __init__(self, colors, defaultColor="black"):
        thresholds = set()
        result = []
        if colors:
            for colorInfo in colors:
                fields = colorInfo.split(None, 1)
                if len(fields) == 2:
                    threshold, color = fields
                    try:
                        threshold = int(threshold)
                    except ValueError:
                        raise ValueError(
                            "color arguments must be given as space-separated "
                            'pairs of "count color" where the count is an '
                            "integer threshold. Your value (%r) was not "
                            "an integer." % threshold
                        )
                    if threshold < 0:
                        raise ValueError(
                            "color arguments must be given as space-separated "
                            'pairs of "count color" where the count is '
                            "non-negative. Your value (%r) is less than 0." % threshold
                        )
                    if threshold in thresholds:
                        raise ValueError(
                            "repeated color argument count (%d)." % threshold
                        )

                    result.append((threshold, color))
                    thresholds.add(threshold)
                else:
                    raise ValueError(
                        "color arguments must be given as space-separated "
                        'pairs of "value color". Your value (%r) does not '
                        "contain a space." % colorInfo
                    )

            result.sort(key=itemgetter(0), reverse=True)

        if not result or result[-1][0] > 0:
            result.append((0, defaultColor))

        self.colors = tuple(result)

    def thresholdToCssName(self, threshold):
        """
        Turn a count threshold into a string that can be used as a CSS
        class name.

        @param threshold: The C{int} threshold.
        @raise ValueError: If the threshold is not an C{int}.
        @return: A C{str} CSS class name.
        """
        return "threshold-%d" % threshold

    def thresholdForCount(self, count):
        """
        Get the best threshold for a specific count.

        @param count: An C{int} count.
        @return: The first C{int} threshold that the given count is at least
            as big as.
        """
        assert count >= 0, "Count (%d) cannot be negative." % count
        for threshold, _ in self.colors:
            if count >= threshold:
                return threshold

        raise ValueError("This should never happen! Last threshold is not 0?")

    def colorForCount(self, count):
        """
        Get the color for a count.

        @param count: An C{int} count.
        @return: The C{str} color for the count.
        """
        assert count >= 0, "Count (%d) cannot be negative." % count
        for threshold, color in self.colors:
            if count >= threshold:
                return color

        raise ValueError("This should never happen! Last threshold is not 0?")



#!/usr/bin/env python

from collections import defaultdict
from math import log10

from dark.aaVars import NAMES
from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions
from dark.summarize import sequenceCategoryLengths


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA on stdin, write a summary of sequence base "
            "categories to stdout. It is currently not possible to "
            "specify the categories on the command line."
        )
    )

    parser.add_argument(
        "--baseType",
        default="nucl",
        choices=("nucl", "prot"),
        help="The type of the bases in the input.",
    )

    parser.add_argument(
        "--minLength",
        default=1,
        type=int,
        help=(
            "If specified, stretches of reads that are less than this "
            "length will not be reported but will be summarized by an "
            "ellipsis."
        ),
    )

    parser.add_argument(
        "--concise",
        action="store_true",
        default=False,
        help="If specified, do not show the individual sequence regions.",
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    if args.baseType == "nucl":
        categories = {
            "A": "nucl",
            "C": "nucl",
            "G": "nucl",
            "T": "nucl",
            "-": "gap",
        }
        default = "ambiguous"
    else:
        categories = {}
        for name in NAMES:
            categories[name] = "aa"
        categories["-"] = "gap"
        default = "ambiguous"

    categoryWidth = max(
        [len(category) for category in categories.values()] + [len(default)]
    )

    minLength = args.minLength
    concise = args.concise

    for index, read in enumerate(reads, start=1):
        counts: dict[str, int] = defaultdict(int)
        readLen = len(read)
        width = int(log10(readLen)) + 1
        if not concise:
            summary: list[str] = []
            append = summary.append
            offset = 1
        for category, count in sequenceCategoryLengths(
            read, categories, defaultCategory=default, minLength=minLength
        ):
            counts[category] += count
            if not concise:
                append(
                    "    %*d %-*s (offset %*d)"
                    % (width, count, categoryWidth, category, width, offset)
                )
                offset += count
        print("%d: %s (length %d)" % (index, read.id, readLen))
        for category in sorted(counts):
            count = counts[category]
            print(
                "  %-*s: %*d (%6.2f%%)"
                % (categoryWidth, category, width, count, count / readLen * 100.0)
            )
        if not concise:
            print("\n".join(summary))



from __future__ import annotations

import re
from math import ceil
from typing import Optional, TYPE_CHECKING
import argparse

if TYPE_CHECKING:
    from dark.reads import Reads
    from dark.titles import TitleAlignments
from dark.simplify import simplifyTitle
from dark.utils import parseRangeExpression


class TitleFilter:
    """
    Provide an acceptance test for sequence titles.

    @param whitelist: If not C{None}, a C{set} of exact titles that are always
        acceptable.
    @param blacklist: If not C{None}, a C{set} of exact titles that are never
        acceptable.
    @param whitelistFile: If not C{None}, a C{str} filename containing lines
        that give exact titles that are always acceptable.
    @param blacklistFile: If not C{None}, a C{str} filename containing lines
        that give exact titles that are never acceptable.
    @param positiveRegex: If not C{None}, a C{str} regex that sequence titles
        must match (case is ignored).
    @param negativeRegex: If not C{None}, a C{str} regex that sequence titles
        must not match (case is ignored).
    @param truncateAfter: A C{str} that titles will be truncated beyond. If
        a truncated title has already been seen, that title will no longer
        be acceptable.
    """

    # The following are return values for the self.accept method.
    REJECT = 0
    WHITELIST_ACCEPT = 1
    DEFAULT_ACCEPT = 2

    def __init__(
        self,
        whitelist: Optional[set[str]] = None,
        blacklist: Optional[set[str]] = None,
        whitelistFile: Optional[str] = None,
        blacklistFile: Optional[str] = None,
        positiveRegex: Optional[str] = None,
        negativeRegex: Optional[str] = None,
        truncateAfter: Optional[str] = None,
    ) -> None:
        whitelist = whitelist or set()
        if whitelistFile:
            with open(whitelistFile) as fp:
                for line in fp:
                    whitelist.add(line.strip())
        self._whitelist = whitelist

        blacklist = blacklist or set()
        if blacklistFile:
            with open(blacklistFile) as fp:
                for line in fp:
                    blacklist.add(line.strip())
        self._blacklist = blacklist

        if truncateAfter is None:
            self._truncated: Optional[dict[str, str]] = None
        else:
            self._truncateAfter = truncateAfter
            self._truncated = {}

        if positiveRegex is None:
            self._positiveRegex = None
        else:
            self._positiveRegex = re.compile(positiveRegex, re.I)

        if negativeRegex is None:
            self._negativeRegex = None
        else:
            self._negativeRegex = re.compile(negativeRegex, re.I)

    def accept(self, title: str) -> int:
        """
        Return a value (see below) to indicate if a title is acceptable (and,
        if so, in what way).

        @param title: A C{str} sequence title.
        @return: An C{int} to indicate an acceptable title or not. This will be

            C{self.REJECT} if the title is unacceptable.
            C{self.WHITELIST_ACCEPT} if the title is whitelisted.
            C{self.DEFAULT_ACCEPT} if the title is acceptable by default.

            These three values are needed so our caller can distinguish between
            the two reasons for acceptance.
        """
        if self._whitelist and title in self._whitelist:
            return self.WHITELIST_ACCEPT

        if self._blacklist and title in self._blacklist:
            return self.REJECT

        # If we have a positive regex but we don't match it, reject.
        if self._positiveRegex and self._positiveRegex.search(title) is None:
            return self.REJECT

        # If we have a negative regex and we do match it, reject.
        if self._negativeRegex and self._negativeRegex.search(title) is not None:
            return self.REJECT

        if self._truncated is not None:
            truncated = simplifyTitle(title, self._truncateAfter)
            if truncated in self._truncated:
                # We've already seen this (truncated) title. Reject unless
                # this is the original title that we truncated to make this
                # entry. That title must continue to be accepted.
                if self._truncated[truncated] == title:
                    return self.DEFAULT_ACCEPT
                else:
                    return self.REJECT
            else:
                self._truncated[truncated] = title

        return self.DEFAULT_ACCEPT


class SequenceFilter:
    """
    Provide an acceptance test for sequences.

    @param whitelist: If not C{None}, a C{set} of exact sequences that are always
        acceptable.
    @param blacklist: If not C{None}, a C{set} of exact sequences that are never
        acceptable.
    @param whitelistFile: If not C{None}, a C{str} filename containing lines
        that give exact sequences that are always acceptable.
    @param blacklistFile: If not C{None}, a C{str} filename containing lines
        that give exact sequences that are never acceptable.
    @param positiveRegex: If not C{None}, a C{str} regex that sequences
        must match (case is ignored).
    @param negativeRegex: If not C{None}, a C{str} regex that sequences
        must not match (case is ignored).
    """

    # The following are return values for the self.accept method.
    REJECT = 0
    WHITELIST_ACCEPT = 1
    DEFAULT_ACCEPT = 2

    def __init__(
        self,
        whitelist: Optional[set[str]] = None,
        blacklist: Optional[set[str]] = None,
        whitelistFile: Optional[str] = None,
        blacklistFile: Optional[str] = None,
        positiveRegex: Optional[str] = None,
        negativeRegex: Optional[str] = None,
    ) -> None:
        whitelist = whitelist or set()
        if whitelistFile:
            with open(whitelistFile) as fp:
                for line in fp:
                    whitelist.add(line.strip())
        self._whitelist = whitelist

        blacklist = blacklist or set()
        if blacklistFile:
            with open(blacklistFile) as fp:
                for line in fp:
                    blacklist.add(line.strip())
        self._blacklist = blacklist

        if positiveRegex is None:
            self._positiveRegex = None
        else:
            self._positiveRegex = re.compile(positiveRegex, re.I)

        if negativeRegex is None:
            self._negativeRegex = None
        else:
            self._negativeRegex = re.compile(negativeRegex, re.I)

    def accept(self, sequence: str) -> int:
        """
        Return a value (see below) to indicate if a sequence is acceptable (and,
        if so, in what way).

        @param sequence: A C{str} sequence.
        @return: An C{int} to indicate an acceptable sequence or not. This will be

            C{self.REJECT} if the sequence is unacceptable.
            C{self.WHITELIST_ACCEPT} if the sequence is whitelisted.
            C{self.DEFAULT_ACCEPT} if the sequence is acceptable by default.

            These three values are needed so our caller can distinguish between
            the two reasons for acceptance.
        """
        if self._whitelist and sequence in self._whitelist:
            return self.WHITELIST_ACCEPT

        if self._blacklist and sequence in self._blacklist:
            return self.REJECT

        # If we have a positive regex but we don't match it, reject.
        if self._positiveRegex and self._positiveRegex.search(sequence) is None:
            return self.REJECT

        # If we have a negative regex and we do match it, reject.
        if self._negativeRegex and self._negativeRegex.search(sequence) is not None:
            return self.REJECT

        return self.DEFAULT_ACCEPT


class ReadSetFilter:
    """
    Provide an acceptance test based on sequence read set.

    @param minNew: The C{float} fraction of its reads by which a new read set
        must differ from all previously seen read sets in order to be
        considered acceptably different.
    """

    def __init__(self, minNew: float) -> None:
        self._minNew = minNew
        self._titles: dict[str, tuple[set[str], list[str]]] = {}

    def accept(self, title: str, titleAlignments: TitleAlignments) -> bool:
        """
        Return C{True} if the read id set in C{titleAlignments} is sufficiently
        different from all previously seen read sets.

        @param title: A C{str} sequence title.
        @param titleAlignments: An instance of L{TitleAlignment}.
        @return: A C{bool} indicating whether a title has an acceptably novel
            read set or not.
        """

        # Sanity check: titles can only be passed once.
        assert title not in self._titles, "Title %r seen multiple times." % title

        readIds = titleAlignments.readIds()
        newReadsRequired = ceil(self._minNew * len(readIds))

        for readSet, invalidatedTitles in self._titles.values():
            if len(readIds - readSet) < newReadsRequired:
                # Add this title to the set of titles invalidated by this
                # previously seen read set.
                invalidatedTitles.append(title)
                return False

        # Remember the new read set and an empty list of invalidated titles.
        self._titles[title] = (readIds, [])

        return True

    def invalidates(self, title: str) -> list[str]:
        """
        Report on which other titles were invalidated by a given title.

        @param title: A C{str} sequence title.
        @return: A C{list} of titles that the passed title invalidated.
        """
        try:
            return self._titles[title][1]
        except KeyError:
            return []


def addFASTAFilteringCommandLineOptions(parser: argparse.ArgumentParser) -> None:
    """
    Add standard FASTA filtering command-line options to an argparse parser.

    These are options that can be used to select or omit entire FASTA records,
    NOT options that change them (for that see
    addFASTAEditingCommandLineOptions).

    @param parser: An C{argparse.ArgumentParser} instance.
    """
    parser.add_argument(
        "--minLength", type=int, metavar="N", help="The minimum sequence length."
    )

    parser.add_argument(
        "--maxLength", type=int, metavar="N", help="The maximum sequence length."
    )

    parser.add_argument(
        "--maxNFraction",
        type=float,
        metavar="N",
        help="The maximum fraction of Ns that can be present in the sequence.",
    )

    parser.add_argument(
        "--whitelist",
        action="append",
        metavar="SEQUENCE-ID",
        help="Sequence titles (ids) that should be whitelisted. May be repeated.",
    )

    parser.add_argument(
        "--blacklist",
        action="append",
        metavar="SEQUENCE-ID",
        help="Sequence titles (ids) that should be blacklisted. May be repeated.",
    )

    parser.add_argument(
        "--whitelistFile",
        metavar="SEQUENCE-ID-FILE",
        help=(
            "The name of a file that contains sequence titles (ids) that "
            "should be whitelisted, one per line"
        ),
    )

    parser.add_argument(
        "--blacklistFile",
        metavar="SEQUENCE-ID-FILE",
        help=(
            "The name of a file that contains sequence titles (ids) that "
            "should be blacklisted, one per line."
        ),
    )

    parser.add_argument(
        "--titleRegex",
        metavar="REGEX",
        help="A regex that sequence titles (ids) must match.",
    )

    parser.add_argument(
        "--negativeTitleRegex",
        metavar="REGEX",
        help="A regex that sequence titles (ids) must not match.",
    )

    parser.add_argument(
        "--sequenceWhitelist",
        action="append",
        metavar="SEQUENCE",
        help="Sequences that should be whitelisted. May be repeated.",
    )

    parser.add_argument(
        "--sequenceBlacklist",
        action="append",
        metavar="SEQUENCE",
        help="Sequences that should be blacklisted. May be repeated.",
    )

    parser.add_argument(
        "--sequenceWhitelistFile",
        metavar="SEQUENCE-FILE",
        help=(
            "The name of a file that contains sequences that should be whitelisted, "
            "one per line."
        ),
    )

    parser.add_argument(
        "--sequenceBlacklistFile",
        metavar="SEQUENCE-FILE",
        help=(
            "The name of a file that contains sequences that should be blacklisted, "
            "one per line."
        ),
    )

    parser.add_argument(
        "--sequenceRegex",
        metavar="REGEX",
        help="A regex that sequences must match.",
    )

    parser.add_argument(
        "--sequenceNegativeRegex",
        metavar="REGEX",
        help="A regex that sequences must not match.",
    )

    # A mutually exclusive group for --keepSequences and --removeSequences.
    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        "--keepSequences",
        metavar="NUMBER,RANGE,...",
        help=(
            "Specify (1-based) ranges of sequence numbers that should be "
            "kept. E.g., --keepSequences 1-3,5 will output just the 1st, "
            "2nd, 3rd, and 5th sequences. All others will be omitted."
        ),
    )

    group.add_argument(
        "--removeSequences",
        metavar="NUMBER,RANGE,...",
        help=(
            "Specify (1-based) ranges of sequence numbers that should be "
            "removed. E.g., --removeSequences 1-3,5 will output all but the "
            "1st, 2nd, 3rd, and 5th sequences. All others will be ouput."
        ),
    )

    parser.add_argument(
        "--head",
        type=int,
        metavar="N",
        help="Only the first N sequences will be printed.",
    )

    # A mutually exclusive group for --removeDuplicates and
    # --removeDuplicatesById
    _removeGroup = parser.add_mutually_exclusive_group()

    _removeGroup.add_argument(
        "--removeDuplicates",
        action="store_true",
        help=(
            "Duplicate reads will be removed, based only on sequence identity. The "
            "first occurrence is kept."
        ),
    )

    _removeGroup.add_argument(
        "--removeDuplicatesById",
        action="store_true",
        help=(
            "Duplicate reads will be removed, based only on read id. The first "
            "occurrence is kept."
        ),
    )

    parser.add_argument(
        "--removeDuplicatesUseMD5",
        action="store_true",
        help=(
            "MD5 sums will be stored instead of the full sequence or read "
            "id when either --removeDuplicates or removeDuplicatesById are "
            "given. One of --removeDuplicates and --removeDuplicatesById "
            "must also be given. Note that this makes duplicate removal "
            "probabilistic. This option can be used to reduce the amount "
            "of RAM consumed during duplicate removal."
        ),
    )

    # See the docstring for dark.reads.Reads.filter for more detail on
    # randomSubset.
    parser.add_argument(
        "--randomSubset",
        type=int,
        metavar="N",
        help=(
            "An integer giving the number of sequences that should be kept. "
            "These will be selected at random."
        ),
    )

    # See the docstring for dark.reads.Reads.filter for more detail on
    # trueLength.
    parser.add_argument(
        "--trueLength",
        type=int,
        metavar="N",
        help=(
            "The number of reads in the FASTA input. Only to be used with "
            "randomSubset"
        ),
    )

    parser.add_argument(
        "--sampleFraction",
        type=float,
        metavar="FRACTION",
        help=(
            "A [0.0, 1.0] C{float} indicating a fraction of the reads that "
            "should be allowed to pass through the filter. The sample size "
            "will only be approximately the product of the sample fraction "
            "and the number of reads. The sample is taken at random."
        ),
    )

    parser.add_argument(
        "--sequenceNumbersFile",
        metavar="FILENAME",
        help=(
            "A file of (1-based) sequence numbers to retain. Numbers must "
            "be one per line."
        ),
    )


def parseFASTAFilteringCommandLineOptions(
    args: argparse.Namespace, reads: Reads
) -> Reads:
    """
    Examine parsed FASTA filtering command-line options and return filtered
    reads.

    @param args: An argparse namespace, as returned by the argparse
        C{parse_args} function.
    @param reads: A C{Reads} instance to filter.
    @return: The filtered C{Reads} instance.
    """
    keepSequences = (
        parseRangeExpression(args.keepSequences, convertToZeroBased=True)
        if args.keepSequences
        else None
    )

    removeSequences = (
        parseRangeExpression(args.removeSequences, convertToZeroBased=True)
        if args.removeSequences
        else None
    )

    return reads.filter(
        minLength=args.minLength,
        maxLength=args.maxLength,
        maxNFraction=args.maxNFraction,
        whitelist=set(args.whitelist) if args.whitelist else None,
        blacklist=set(args.blacklist) if args.blacklist else None,
        whitelistFile=args.whitelistFile,
        blacklistFile=args.blacklistFile,
        titleRegex=args.titleRegex,
        negativeTitleRegex=args.negativeTitleRegex,
        sequenceWhitelist=set(args.sequenceWhitelist)
        if args.sequenceWhitelist
        else None,
        sequenceBlacklist=set(args.sequenceBlacklist)
        if args.sequenceBlacklist
        else None,
        sequenceWhitelistFile=args.sequenceWhitelistFile,
        sequenceBlacklistFile=args.sequenceBlacklistFile,
        sequenceRegex=args.sequenceRegex,
        sequenceNegativeRegex=args.sequenceNegativeRegex,
        keepSequences=keepSequences,
        removeSequences=removeSequences,
        head=args.head,
        removeDuplicates=args.removeDuplicates,
        removeDuplicatesById=args.removeDuplicatesById,
        removeDuplicatesUseMD5=args.removeDuplicatesUseMD5,
        randomSubset=args.randomSubset,
        trueLength=args.trueLength,
        sampleFraction=args.sampleFraction,
        sequenceNumbersFile=args.sequenceNumbersFile,
    )


def addFASTAEditingCommandLineOptions(parser: argparse.ArgumentParser) -> None:
    """
    Add standard FASTA editing command-line options to an argparse parser.

    These are options that can be used to alter FASTA records, NOT options
    that simply select or reject those things (for those see
    addFASTAFilteringCommandLineOptions).

    @param parser: An C{argparse.ArgumentParser} instance.
    """
    # A mutually exclusive group for --keepSites, --keepSitesFile,
    # --removeSites, and --removeSitesFile.
    group = parser.add_mutually_exclusive_group()

    # In the 4 options below, the 'indices' alternate names are kept for
    # backwards compatibility.
    group.add_argument(
        "--keepSites",
        "--keepIndices",
        help=(
            "Specify 1-based sequence sites to keep. All other sites will "
            "be removed. The sites must be given in the form e.g., "
            "24,100-200,260. Note that the requested sites will be taken "
            "from the input sequences in order, not in the order given by "
            "--keepSites. I.e., --keepSites 5,8-10 will get you the same "
            "result as --keepSites 8-10,5."
        ),
    )

    group.add_argument(
        "--keepSitesFile",
        "--keepIndicesFile",
        help=(
            "Specify a file containing 1-based sites to keep. All other "
            "sequence sites will be removed. Lines in the file must be "
            "given in the form e.g., 24,100-200,260. See --keepSites for "
            "more detail."
        ),
    )

    group.add_argument(
        "--removeSites",
        "--removeIndices",
        help=(
            "Specify 1-based sites to remove. All other sequence sites will "
            "be kept. The sites must be given in the form e.g., "
            "24,100-200,260. See --keepSites for more detail."
        ),
    )

    group.add_argument(
        "--removeSitesFile",
        "--removeIndicesFile",
        help=(
            "Specify a file containing 1-based sites to remove. All other "
            "sequence sites will be kept. Lines in the file must be given "
            "in the form e.g., 24,100-200,260. See --keepSites for more "
            "detail."
        ),
    )

    parser.add_argument(
        "--removeGaps",
        action="store_true",
        help="If True, gap ('-') characters in sequences will be removed.",
    )

    parser.add_argument(
        "--truncateTitlesAfter",
        help=(
            "A string that sequence titles (ids) will be truncated beyond. "
            "If the truncated version of a title has already been seen, "
            "that title will be skipped."
        ),
    )

    parser.add_argument(
        "--removeDescriptions",
        action="store_true",
        help=(
            "Read id descriptions will be removed. The "
            "description is the part of a sequence id after the "
            "first whitespace (if any)."
        ),
    )

    parser.add_argument(
        "--idLambda",
        metavar="LAMBDA-FUNCTION",
        help=(
            "A one-argument function taking and returning a read id. "
            "E.g., --idLambda \"lambda id: id.split('_')[0]\" or "
            '--idLambda "lambda id: id[:10]". If the function returns None, '
            "the read will be filtered out."
        ),
    )

    parser.add_argument(
        "--readLambda",
        metavar="LAMBDA-FUNCTION",
        help=(
            "A one-argument function taking and returning a read. "
            "E.g., --readLambda \"lambda r: Read(r.id.split('_')[0], "
            "r.sequence.strip('-'))\". Make sure to also modify the "
            "quality string if you change the length of a FASTQ sequence. "
            "If the function returns None, the read will be filtered out. "
            "The function will be passed to eval with the dark.reads "
            "classes Read, DNARead, AARead, etc. all in scope."
        ),
    )

    parser.add_argument(
        "--reverse",
        action="store_true",
        help="Reverse the sequences. Note that this is NOT reverse complementing.",
    )

    parser.add_argument(
        "--reverseComplement",
        action="store_true",
        help="Reverse complement the sequences.",
    )

    parser.add_argument(
        "--rotate",
        type=int,
        metavar="N",
        help=(
            "Rotate sequences. A positive integer rotates the sequence to the right, "
            "negative to the left"
        ),
    )

    seqCaseGroup = parser.add_mutually_exclusive_group()

    seqCaseGroup.add_argument(
        "--upper",
        action="store_true",
        help="Convert sequences to uppercase.",
    )

    seqCaseGroup.add_argument(
        "--lower",
        action="store_true",
        help="Convert sequences to lowercase.",
    )

    idCaseGroup = parser.add_mutually_exclusive_group()

    idCaseGroup.add_argument(
        "--upperId",
        action="store_true",
        help="Convert sequence IDs to uppercase.",
    )

    idCaseGroup.add_argument(
        "--lowerId",
        action="store_true",
        help="Convert sequence IDs to lowercase.",
    )


def parseFASTAEditingCommandLineOptions(
    args: argparse.Namespace, reads: Reads
) -> Reads:
    """
    Examine parsed FASTA editing command-line options and return information
    about kept sites and sequences.

    @param args: An argparse namespace, as returned by the argparse
        C{parse_args} function.
    @param reads: A C{Reads} instance to filter.
    @return: The filtered C{Reads} instance.
    """
    removeGaps = args.removeGaps
    removeDescriptions = args.removeDescriptions
    truncateTitlesAfter = args.truncateTitlesAfter
    keepSites = (
        parseRangeExpression(args.keepSites, convertToZeroBased=True)
        if args.keepSites
        else None
    )

    if args.keepSitesFile:
        keepSites = keepSites or set()
        with open(args.keepSitesFile) as fp:
            for lineNumber, line in enumerate(fp):
                try:
                    keepSites.update(
                        parseRangeExpression(line, convertToZeroBased=True)
                    )
                except ValueError as e:
                    raise ValueError(
                        "Keep sites file %r line %d could not be parsed: "
                        "%s" % (args.keepSitesFile, lineNumber, e)
                    )

    removeSites = (
        parseRangeExpression(args.removeSites, convertToZeroBased=True)
        if args.removeSites
        else None
    )

    if args.removeSitesFile:
        removeSites = removeSites or set()
        with open(args.removeSitesFile) as fp:
            for lineNumber, line in enumerate(fp):
                try:
                    removeSites.update(
                        parseRangeExpression(line, convertToZeroBased=True)
                    )
                except ValueError as e:
                    raise ValueError(
                        "Remove sites file %r line %d parse error: %s"
                        % (args.removeSitesFile, lineNumber, e)
                    )

    return reads.filter(
        removeGaps=removeGaps,
        truncateTitlesAfter=truncateTitlesAfter,
        removeDescriptions=removeDescriptions,
        idLambda=args.idLambda,
        readLambda=args.readLambda,
        keepSites=keepSites,
        removeSites=removeSites,
        reverse=args.reverse,
        reverseComplement=args.reverseComplement,
        upper=args.upper,
        lower=args.lower,
        upperId=args.upperId,
        lowerId=args.lowerId,
        rotate=args.rotate,
    )



import numpy as np

START_CODONS = set(["ATG"])
STOP_CODONS = set(["TAA", "TAG", "TGA"])


def findCodons(seq, codons):
    """
    Find all instances of the codons in 'codons' in the given sequence.

    seq: A Bio.Seq.Seq instance.
    codons: A set of codon strings.

    Return: a generator yielding matching codon offsets.
    """
    seqLen = len(seq)
    start = 0
    while start < seqLen:
        triplet = str(seq[start : start + 3])
        if triplet in codons:
            yield start
        start = start + 3


def addORFs(fig, seq, minX, maxX, offsetAdjuster):
    """
    fig is a matplotlib figure.
    seq is a Bio.Seq.Seq.
    minX: the smallest x coordinate.
    maxX: the largest x coordinate.
    featureEndpoints: an array of features as returned by addFeatures (may be
        empty).
    offsetAdjuster: a function to adjust feature X axis offsets for plotting.
    """
    for frame in range(3):
        target = seq[frame:]
        for codons, codonType, color in (
            (START_CODONS, "start", "green"),
            (STOP_CODONS, "stop", "red"),
        ):
            offsets = list(map(offsetAdjuster, findCodons(target, codons)))
            if offsets:
                fig.plot(
                    offsets,
                    np.tile(frame, len(offsets)),
                    marker=".",
                    markersize=4,
                    color=color,
                    linestyle="None",
                )

    fig.axis([minX, maxX, -1, 3])
    fig.set_yticks(np.arange(3))
    fig.set_ylabel("Frame", fontsize=17)
    fig.set_title(
        "Subject start (%s) and stop (%s) codons"
        % (", ".join(sorted(START_CODONS)), ", ".join(sorted(STOP_CODONS))),
        fontsize=20,
    )


def addReversedORFs(fig, seq, minX, maxX, offsetAdjuster):
    """
    fig is a matplotlib figure.
    seq is a Bio.Seq.Seq (the reverse complement of the sequence we're
        plotting against).
    minX: the smallest x coordinate.
    maxX: the largest x coordinate.
    offsetAdjuster: a function to adjust feature X axis offsets for plotting.
    """
    for frame in range(3):
        target = seq[frame:]
        for codons, codonType, color in (
            (START_CODONS, "start", "green"),
            (STOP_CODONS, "stop", "red"),
        ):
            offsets = [
                maxX - offsetAdjuster(offset) for offset in findCodons(target, codons)
            ]
            if offsets:
                fig.plot(
                    offsets,
                    np.tile(frame, len(offsets)),
                    marker=".",
                    markersize=4,
                    color=color,
                    linestyle="None",
                )

    fig.axis([minX, maxX, -1, 3])
    fig.set_yticks(np.arange(3))
    fig.set_ylabel("Frame", fontsize=17)
    fig.set_title(
        "Reversed subject start (%s) and stop (%s) codons"
        % (", ".join(sorted(START_CODONS)), ", ".join(sorted(STOP_CODONS))),
        fontsize=20,
    )



#!/usr/bin/env python

import sys
import argparse

from dark.idutils import ids


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "spec",
        nargs="+",
        help=(
            "One or two arguments to specify the start and (optionally) the end "
            "range of ids to create."
        ),
        metavar="ID",
    )

    parser.add_argument(
        "--prefix",
        help=(
            "The id prefix. If this is not given and two arguments are supplied, "
            "their longest common prefix will be used."
        ),
    )

    parser.add_argument(
        "--sep", default=" ", help="The string to print between successive ids."
    )

    parser.add_argument(
        "-0",
        "--zeroes",
        action="store_true",
        help="Print numeric suffixes with leading zeroes.",
    )

    parser.add_argument(
        "-n",
        "--count",
        dest="maxResults",
        type=int,
        help="The number of items to generate.",
    )

    args = parser.parse_args()

    if len(args.spec) == 1:
        start, end = args.spec[0], None
    elif len(args.spec) == 2:
        start, end = args.spec
    else:
        print("I only know how to process one or two arguments.", file=sys.stderr)
        sys.exit(1)

    sep = {
        r"\t": "\t",
        r"\n": "\n",
    }.get(args.sep, args.sep)

    # Iterate through the yielded results instead of trying to compute them
    # all at once and joining the resulting list.
    first = True

    for id_ in ids(start, end, args.prefix, sep, args.zeroes, args.maxResults):
        if not first:
            print(sep, end="")
        print(f"{id_}", end="")
        first = False

    if not first:
        print()


if __name__ == "__main__":
    main()



from unittest import TestCase

from dark.distance import levenshtein


class TestLevenshtein(TestCase):
    """
    Tests for the dark.simplify.levenshtein function.
    """

    def testIdentical(self):
        """
        Two identical strings must have distance zero.
        """
        self.assertEqual(0, levenshtein("BLAH", "BLAH"))

    def testMutation(self):
        """
        Test a single character results in a distance of 1.
        """
        self.assertEqual(1, levenshtein("ACGTACACACG", "ACGTACACACT"))

    def testInsert(self):
        """
        Test a string insertion that results in a distance of 2.
        """
        self.assertEqual(2, levenshtein("AGTACACACTG", "ACGTACACACT"))



import multiprocessing
from tempfile import mkdtemp
import os
from os.path import join
import requests

from dark.process import Executor
from dark.sam import samfile as openSamfile


class Bowtie2:
    """
    Run Bowtie2.
    """

    def __init__(
        self,
        executor=None,
        threads=None,
        verboseFp=None,
        dryRun=False,
        reference=None,
        tempdir=None,
        tmpChmod=None,
    ):
        self._executor = executor or Executor(dryRun)
        if dryRun:
            self.tempdir = tempdir or "/tmp/xxx"
        else:
            self.tempdir = mkdtemp(prefix="bt2-", dir=tempdir)
            if tmpChmod:
                self._executor.execute(f"chmod {tmpChmod} {self.tempdir}")
        self._samFile = join(self.tempdir, "result.sam")
        self._bamFile = join(self.tempdir, "result.bam")
        self._indexFile = join(self.tempdir, "index")
        self._verboseFp = verboseFp
        self._indexCalled = self._samExists = self._bamExists = False
        self.nThreads = threads or multiprocessing.cpu_count()
        self._reference = reference

    def buildIndex(self, index):
        """
        Find or make a Bowtie2 index.
        """
        if os.path.exists(index):
            # Check if this is a pre-existing bowtie2 index. Look for and
            # remove a bowtie2 suffix, if any (otherwise bowtie2 will
            # complain). We do things this way to allow a user to use TAB
            # completion on the command line to give them the full path to
            # any bowtie index file.
            for suffix in ("1.bt2 2.bt2 3.bt2 4.bt2 rev.1.bt2 rev.2.bt2").split():
                suffix = "." + suffix
                if index.endswith(suffix):
                    self._indexFile = index[: -len(suffix)]
                    self._report(
                        "Using pre-existing Bowtie2 index %r." % self._indexFile
                    )
                    break
            else:
                # Assume a FASTA file and make an index.
                self._indexFile = self._makeIndexFromFastaFile(index)
        else:
            # Not a filename. So either the start of the path to a bowtie2
            # index or else an accession number.
            if os.path.exists(index + ".1.bt2"):
                self._report("Using pre-existing Bowtie2 index %r." % index)
                self._indexFile = index
            else:
                # Assume an accession number.
                self._indexFile = self._makeIndexFromAccession(index)

        self._indexCalled = True

    def align(
        self,
        bowtie2Args="--no-unal",
        fastq1=None,
        fastq2=None,
        threads=None,
        discardSAM=False,
        readGroup="orig",
        sampleName="orig",
    ):
        """
        Run Bowtie2.
        """
        if not self._indexCalled:
            raise ValueError("buildIndex() has not yet been called.")

        self._report("Aligning with Bowtie2.")

        nThreads = threads or self.nThreads
        samFile = "/dev/null" if discardSAM else self._samFile

        if fastq1 and fastq2:
            self._executor.execute(
                "bowtie2 %s --threads %d --rg-id '%s' --rg 'SM:%s' -x '%s' "
                "-1 '%s' -2 '%s' > '%s'"
                % (
                    bowtie2Args,
                    nThreads,
                    readGroup,
                    sampleName,
                    self._indexFile,
                    fastq1,
                    fastq2,
                    samFile,
                )
            )
        elif fastq1:
            self._executor.execute(
                "bowtie2 %s --threads %d --rg-id '%s' --rg 'SM:%s' -x '%s' "
                "-U '%s' > '%s'"
                % (
                    bowtie2Args,
                    nThreads,
                    readGroup,
                    sampleName,
                    self._indexFile,
                    fastq1,
                    samFile,
                )
            )
        else:
            raise ValueError("At least fastq1 must be passed.")

        self._samExists = True

    def _report(self, mesg):
        """
        Print a progress message.
        """
        if self._verboseFp:
            print(mesg, file=self._verboseFp)

    def makeBAM(self, samtoolsViewArgs=""):
        """
        Convert the SAM to BAM.
        """
        # The following will raise if there is no SAM file.
        self._SAMorBAM()
        self._report("Converting SAM to BAM.")
        self._executor.execute(
            "samtools view -b %s '%s' > '%s'"
            % (samtoolsViewArgs, self._samFile, self._bamFile)
        )
        self._bamExists = True

    def _SAMorBAM(self):
        """
        Do we have a SAM or BAM file available?
        """
        if self._bamExists:
            return "BAM"
        elif self._samExists:
            return "SAM"
        else:
            raise ValueError("bowtie2() has not yet been called.")

    def outputFile(self):
        """
        The name of the output file.
        """
        if self._bamExists:
            return self._bamFile
        elif self._samExists:
            return self._samFile
        else:
            raise ValueError("bowtie2() has not yet been called.")

    def indexBAM(self):
        """
        Index the BAM file.
        """
        which = self._SAMorBAM()

        if which != "BAM":
            raise ValueError("makeBAM() has not yet been called.")

        self._report("Indexing BAM.")
        self._executor.execute("samtools index '%s'" % self._bamFile)

    def mappedAndUnmappedCounts(self) -> tuple[int, int]:
        """
        Return the number of mapped and unmapped reads.

        @return: A 2-C{tuple} of C{int}s, giving the number of mapped and unmapped
            reads.
        """
        filename = self._bamFile if self._SAMorBAM() == "BAM" else self._samFile
        with openSamfile(filename) as samfile:
            return samfile.mapped, samfile.unmapped

    def sort(self, byName=False):
        """
        Sort the BAM or SAM.
        """
        which = self._SAMorBAM()
        self._report("Sorting %s (by %s)." % (which, "name" if byName else "coord"))
        inFile = self._bamFile if which == "BAM" else self._samFile
        sortedFile = join(self.tempdir, "result-sorted." + which.lower())
        self._executor.execute(
            "samtools sort %s'%s' > '%s'"
            % ("-n " if byName else "", inFile, sortedFile)
        )
        self._executor.execute("mv '%s' '%s'" % (sortedFile, inFile))

    def removePrimers(self, bedFile):
        """
        Removes primers specified in the bed file
        """
        which = self._SAMorBAM()

        if which != "BAM":
            raise ValueError("makeBAM() has not yet been called.")

        self._report("removing primers specified in %s" % bedFile)
        tempTrimmedBamPrefix = "%s.trimmed" % self._bamFile
        self._executor.execute(
            "ivar trim -b '%s' -p '%s' -i '%s' -q 20 -m 30 -s 4 -e"
            % (bedFile, tempTrimmedBamPrefix, self._bamFile)
        )
        self._executor.execute(
            "mv '%s'.bam '%s'" % (tempTrimmedBamPrefix, self._bamFile)
        )

    def markDuplicatesPicard(self, picardFile):
        """
        Use Picard to mark duplicates.
        """
        which = self._SAMorBAM()
        self._report("Marking duplicates with Picard.")

        inFile = self._bamFile if which == "BAM" else self._samFile
        tempFile = join(self.tempdir, "picard-duplicates." + which.lower())
        tempErrFile = join(self.tempdir, "picard.errs")

        self._executor.execute(
            "java -Xmn2g -Xms2g -Xmx2g -jar %s "
            "MarkDuplicates I='%s' O='%s' M=/dev/null >'%s' 2>&1"
            % (picardFile, inFile, tempFile, tempErrFile)
        )

        self._executor.execute("mv '%s' '%s'" % (tempFile, inFile))

    def markDuplicatesGATK(self, threads=None):
        """
        Use GATK to mark duplicates.
        """
        nThreads = threads or self.nThreads
        which = self._SAMorBAM()
        self._report("Marking duplicates with GATK.")

        inFile = self._bamFile if which == "BAM" else self._samFile
        tempFile = join(self.tempdir, "gatk-duplicates." + which.lower())

        self._executor.execute(
            "gatk MarkDuplicatesSpark "
            "-I '%s' -O '%s' --conf spark.executor.cores=%d"
            % (inFile, tempFile, nThreads)
        )

        self._executor.execute("mv '%s' '%s'" % (tempFile, inFile))

    def callHaplotypesGATK(self, picardJar, vcfFile=None, referenceFasta=None):
        """
        Use GATK to call haplotypes.
        """
        which = self._SAMorBAM()
        self._report("Calling haplotypes with GATK.")

        inFile = self._bamFile if which == "BAM" else self._samFile
        vcfFile = vcfFile or join(self.tempdir, "output.vcf.gz")

        if referenceFasta is None:
            if self._reference:
                self._report("Using %s as a reference." % self._reference)
                referenceFasta = self._reference
            else:
                raise ValueError(
                    "No reference was passed, given to the "
                    "Bowtie2 __init__, or used in buildIndex."
                )

        indexFile = referenceFasta + ".fai"
        if os.path.exists(indexFile):
            removeIndex = False
        else:
            removeIndex = True
            self._executor.execute("samtools faidx '%s'" % referenceFasta)

        if referenceFasta.lower().endswith(".fasta"):
            dictFile = referenceFasta[: -len(".fasta")] + ".dict"
        else:
            dictFile = referenceFasta + ".dict"

        if os.path.exists(dictFile):
            removeDict = False
        else:
            removeDict = True
            self._executor.execute(
                "java -jar '%s' CreateSequenceDictionary R='%s' O='%s'"
                % (picardJar, referenceFasta, dictFile)
            )

        self._executor.execute(
            "gatk --java-options -Xmx4g HaplotypeCaller "
            "--reference '%s' "
            "--input '%s' "
            "--output '%s' "
            "--sample-ploidy 1 "
            "--dont-use-soft-clipped-bases true "
            "-ERC GVCF" % (referenceFasta, inFile, vcfFile)
        )

        if removeIndex:
            self._executor.execute("rm '%s'" % indexFile)

        if removeDict:
            self._executor.execute("rm '%s'" % dictFile)

    def callHaplotypesBcftools(self, vcfFile=None, referenceFasta=None):
        """
        Use bcftools call to call haplotypes.
        """
        which = self._SAMorBAM()
        self._report("Calling haplotypes with bcftools call.")

        inFile = self._bamFile if which == "BAM" else self._samFile
        vcfFile = vcfFile or join(self.tempdir, "output.vcf.gz")

        if referenceFasta is None:
            if self._reference:
                self._report("Using %s as a reference." % self._reference)
                referenceFasta = self._reference
            else:
                raise ValueError(
                    "No reference was passed, given to the "
                    "Bowtie2 __init__, or used in buildIndex."
                )

        self._executor.execute(
            'bcftools mpileup --max-depth 5000 -Ou -f "%s" "%s" | '
            'bcftools call --ploidy 1 -mv -Oz -o "%s"'
            % (referenceFasta, inFile, vcfFile)
        )

        self._executor.execute(f"bcftools index --force {vcfFile!r}")

    def removeDuplicates(self):
        """
        Use samtools to remove marked duplicates.
        """
        which = self._SAMorBAM()
        self._report("Removing marked duplicates.")

        inFile = self._bamFile if which == "BAM" else self._samFile
        tempFile = join(self.tempdir, "non-duplicates." + which.lower())

        # See the comment on the testRemoveDuplicates test in
        # ../test/test_bowtie2.py regarding there being two spaces in the
        # samtools command below if we are not producing BAM.  If you
        # change spacing here you will also need to change that test.
        bArg = "-b" if which == "BAM" else ""
        self._executor.execute(
            "samtools view %s -F 1024 '%s' > '%s'" % (bArg, inFile, tempFile)
        )
        self._executor.execute("mv '%s' '%s'" % (tempFile, inFile))

    def _makeIndexFromFastaFile(self, fastaFilename):
        index = join(self.tempdir, "index")

        self._report("Building Bowtie2 index from %s." % fastaFilename)
        self._executor.execute(
            "bowtie2-build --quiet '%s' '%s'" % (fastaFilename, index)
        )

        if self._reference is None:
            self._report("Will use this FASTA file as the reference (if needed).")
            self._reference = fastaFilename

        return index

    def _makeIndexFromAccession(self, accessionId):
        fastaFilename = join(self.tempdir, accessionId + ".fasta")
        index = join(self.tempdir, "index")

        self._report("Downloading FASTA for accession %s from NCBI." % accessionId)

        URL = (
            "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?"
            "db=nucleotide&id=%s&rettype=fasta&retmode=text" % accessionId
        )

        if not self._executor.dryRun:
            with open(fastaFilename, "w") as fp:
                print(requests.get(URL).text.rstrip("\n"), file=fp)

        self._report("Building bowtie2 index for %s." % accessionId)
        self._executor.execute(
            "bowtie2-build --quiet '%s' '%s'" % (fastaFilename, index)
        )

        return index

    def close(self):
        """
        Clean up.
        """
        self._executor.execute("rm -r '%s'" % self.tempdir)



#!/usr/bin/env python

import sys
import argparse
import dendropy

# See https://jeetsukumaran.github.io/DendroPy/schemas/index.html for the available
# formats.
FORMATS = {"newick", "nexml", "nexus", "phylip"}


def parseArgs():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="""Special-purpose script to add support labels to nodes in
trees produced by `iqtree2` when (if not run with `-keep-ident`) it adds nodes and tips
for identical sequences. iqtree2 currently does not put a support label on (the edge
leading to) the tip in the original processingor onto the nodes introduced by adding
tips for the identical sequences.  This is described at
https://github.com/iqtree/iqtree2/issues/343 The adjusted tree is written to standard
output in the same format as the input tree (as specified by --format argument).""",
    )

    parser.add_argument(
        "treeFile",
        help="The tree file to examine.",
    )

    parser.add_argument(
        "--support",
        default="100",
        help="The support label to add to unlabeled internal nodes",
    )

    parser.add_argument(
        "--format",
        default="newick",
        choices=sorted(FORMATS),
        help="The input (and resulting output) tree format.",
    )

    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Print information about the number of nodes to which a label was added.",
    )

    return parser.parse_args()


def main() -> None:
    """
    Add a label (a support value) to internal nodes with no label and print the
    resultant tree to standard output.
    """
    args = parseArgs()

    kwargs = (
        dict(preserve_underscores=True) if args.format in {"newick", "nexus"} else {}
    )
    with open(args.treeFile) as fp:
        tree = dendropy.Tree.get(file=fp, schema=args.format, **kwargs)

    changes = 0

    for node in tree.preorder_internal_node_iter():
        if node.label is None:
            if node is not tree.seed_node:
                childLengths = [child.edge.length for child in node.child_nodes()]
                if childLengths != [0.0, 0.0]:
                    exit(
                        "Internal node found with edges to its children that are not "
                        f"all zero length (but instead, {', '.join(childLengths)}. "
                        "Are you sure this tree was produced by iqtree2?"
                    )
                node.label = args.support
                changes += 1

    if args.verbose:
        if changes:
            s = "" if changes == 1 else "s"
            print(
                f"Added label {args.support!r} to {changes} node{s}.",
                file=sys.stderr,
            )
        else:
            print(
                "No unlabeled nodes were found. The input and output trees will be "
                "identical.",
                file=sys.stderr,
            )

    tree.write(file=sys.stdout, schema=args.format)


if __name__ == "__main__":
    main()



from unittest import TestCase
import warnings

from dark.genbank import GenomeRanges


class TestErrors(TestCase):
    """
    Test incorrect ways of instantitating the GenomeRanges class.
    """

    def testEmptyString(self):
        """
        The GenomeRanges class must raise a ValueError when passed an range
        empty string.
        """
        error = (
            r'^Could not parse GenBank range string ""\. '
            r'Range "" does not end with \]\(\+\) or \]\(-\)\.$'
        )
        self.assertRaisesRegex(ValueError, error, GenomeRanges, "")

    def testUnparseableString(self):
        """
        The GenomeRanges class must raise a ValueError when passed an
        unparseable string.
        """
        error = (
            r'^Could not parse GenBank range string "x"\. '
            r'Range "x" does not end with \]\(\+\) or \]\(-\)\.$'
        )
        self.assertRaisesRegex(ValueError, error, GenomeRanges, "x")

    def testNoComplementIndicator(self):
        """
        The GenomeRanges class must raise a ValueError when passed an a range
        that does not end with ](+] or ](-).
        """
        error = (
            r'^Could not parse GenBank range string "\[33:40\]"\. '
            r'Range "\[33:40\]" does not end with \]\(\+\) or \]\(-\)\.$'
        )
        self.assertRaisesRegex(ValueError, error, GenomeRanges, "[33:40]")

    def testNoLeadingSquareBracket(self):
        """
        The GenomeRanges class must raise a ValueError when passed an a range
        that does not start with [.
        """
        error = (
            r'^Could not parse GenBank range string "33:40\]\(-\)"\. '
            r'Range "33:40\]\(-\)" does not start with "\["\.$'
        )
        self.assertRaisesRegex(ValueError, error, GenomeRanges, "33:40](-)")

    def testJoinWithNoRange(self):
        """
        The GenomeRanges class must raise a ValueError when passed a join
        directive containing no ranges.
        """
        error = (
            r'^Could not parse GenBank range string "join{}"\. '
            r"join{} can only be used with multiple ranges\.$"
        )
        self.assertRaisesRegex(ValueError, error, GenomeRanges, "join{}")

    def testDecreasingOrder(self):
        """
        The GenomeRanges class must raise a ValueError when passed a string in
        which the start, stop values decrease.
        """
        error = (
            r'^Could not parse GenBank range string "\[7:4\]\(\+\)"\. '
            r"Offset values \(7, 4\) cannot decrease\.$"
        )
        self.assertRaisesRegex(ValueError, error, GenomeRanges, "[7:4](+)")

    def testContiguousRanges(self):
        """
        The GenomeRanges class must raise a ValueError if passed two ranges
        that are contiguous and which have the same (-) direction.
        """
        message = "Contiguous GenBank ranges detected: [3:5] followed by [5:7]."
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            GenomeRanges("join{[3:5](-), [5:7](-)}")
            self.assertEqual(1, len(w))
            self.assertEqual(message, str(w[0].message))

    def testContiguousRangesComplementStrand(self):
        """
        The GenomeRanges class must raise a ValueError if passed two ranges
        that are contiguous and which have the same (+) direction.
        """
        message = "Contiguous GenBank ranges detected: [3:5] followed by [5:7]."
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            GenomeRanges("join{[3:5](+), [5:7](+)}")
            self.assertEqual(1, len(w))
            self.assertEqual(message, str(w[0].message))

    def testTwoNakedRanges(self):
        """
        Two ranges with no join qualifier must result in a ValueError.
        """
        error = (
            r"^Could not parse GenBank range string "
            r'"\[3:5\(\+\), 7:9\(-\)\]"\. '
            r"Multiple ranges must be wrapped in join{}\.$"
        )
        self.assertRaisesRegex(ValueError, error, GenomeRanges, "[3:5(+), 7:9(-)]")

    def testOneJoinedRange(self):
        """
        A single range with a join() must result in a ValueError.
        """
        error = (
            r"^Could not parse GenBank range string "
            r'"join{\[3:5\]\(-\)}"\. '
            r"join{} can only be used with multiple ranges\.$"
        )
        self.assertRaisesRegex(ValueError, error, GenomeRanges, "join{[3:5](-)}")


class TestRanges(TestCase):
    """
    Test the ranges attribute of the GenomeRanges class.
    """

    def testOneNakedRangePositive(self):
        """
        A single range on the positive strand must result in the the expected
        ranges value stored.
        """
        gr = GenomeRanges("[3:5](+)")
        self.assertEqual(((3, 5, True),), gr.ranges)

    def testOneNakedRangeNegative(self):
        """
        A single range on the negative strand must result in the the expected
        ranges value stored.
        """
        gr = GenomeRanges("[3:5](-)")
        self.assertEqual(((3, 5, False),), gr.ranges)

    def testTwoJoinedRanges(self):
        """
        Two joined ranges must return the expected result.
        """
        gr = GenomeRanges("join{[3:5](+), [7:9](-)}")
        self.assertEqual(((3, 5, True), (7, 9, False)), gr.ranges)

    def testThreeJoinedRanges(self):
        """
        Three joined ranges must return the expected result.
        """
        gr = GenomeRanges("join{[3:5](+), [7:9](-), [17:19](-)}")
        self.assertEqual(((3, 5, True), (7, 9, False), (17, 19, False)), gr.ranges)

    def testTwoJoinedContiguousRangesMismatchedStrands(self):
        """
        Two joined ranges that are contiguous but not on the same strand must
        return the expected unmerged (two-range) result.
        """
        gr = GenomeRanges("join{[3:5](-), [5:9](+)}")
        self.assertEqual(((3, 5, False), (5, 9, True)), gr.ranges)

    def testTwoJoinedContiguousRangesComplementStrand(self):
        """
        Two joined ranges that are contiguous on the same (complement) strand
        must return the expected (single-range) result.
        """
        message = "Contiguous GenBank ranges detected: [3:5] followed by [5:9]."
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            gr = GenomeRanges("join{[3:5](-), [5:9](-)}")
            self.assertEqual(((3, 9, False),), gr.ranges)
            self.assertEqual(1, len(w))
            self.assertEqual(message, str(w[0].message))

    def testTwoJoinedContiguousRanges(self):
        """
        Two joined ranges that are contiguous on the same strand must return
        the expected (single-range) result.
        """
        message = "Contiguous GenBank ranges detected: [3:5] followed by [5:9]."
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            gr = GenomeRanges("join{[3:5](+), [5:9](+)}")
            self.assertEqual(((3, 9, True),), gr.ranges)
            self.assertEqual(1, len(w))
            self.assertEqual(message, str(w[0].message))

    def testThreeJoinedContiguousRanges(self):
        """
        Three joined ranges that are contiguous must return the expected
        (single-range) result.
        """
        message1 = "Contiguous GenBank ranges detected: [3:5] followed by [5:7]."
        # Note that the second warning message has a range that doesn't
        # correspond to any of the ranges in the GenomeRanges
        # initialization string. That's because by the time the second
        # warning is issued the first two ranges ([3:5] and [5:7]) have
        # been merged into one ([3:7]).
        message2 = "Contiguous GenBank ranges detected: [3:7] followed by [7:9]."
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            gr = GenomeRanges("join{[3:5](-), [5:7](-), [7:9](-)}")
            self.assertEqual(((3, 9, False),), gr.ranges)
            self.assertEqual(2, len(w))
            self.assertEqual(message1, str(w[0].message))
            self.assertEqual(message2, str(w[1].message))

    def testTwoJoinedContiguousRangesInTheMiddleOfFourRanges(self):
        """
        Two joined ranges that are contiguous must return the expected
        result when there are non-contiguous ranges surrounding them.
        """
        message = "Contiguous GenBank ranges detected: [3:5] followed by [5:9]."
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            gr = GenomeRanges("join{[0:2](-), [3:5](-), [5:9](-), [12:15](-)}")
            self.assertEqual(((0, 2, False), (3, 9, False), (12, 15, False)), gr.ranges)
            self.assertEqual(1, len(w))
            self.assertEqual(message, str(w[0].message))


class TestCircular(TestCase):
    """
    Test the GenomeRanges class ranges method.
    """

    def testOneRange(self):
        """
        The circular method must return False when given only a single range
        tuple that is fully contained within the genome.
        """
        self.assertFalse(GenomeRanges("[20:40](+)").circular(100))

    def testOneRangeEndingAtGenomeEnd(self):
        """
        The circular method must return False when given only a single range
        tuple that ends at the end of the genome.
        """
        self.assertFalse(GenomeRanges("[20:40](+)").circular(40))

    def testOneRangeSpanningTheWholeGenome(self):
        """
        The circular method must return False when given only a single range
        tuple that ends at the end of the genome.
        """
        self.assertFalse(GenomeRanges("[0:40](+)").circular(40))

    def testTwoRangesThatAreCircular(self):
        """
        The circular method must return True when given two ranges that span
        the end of the genome.
        """
        self.assertTrue(GenomeRanges("join{[20:40](+), [0:10](+)}").circular(40))

    def testThreeRangesThatAreCircular(self):
        """
        The circular method must return True when given three ranges that span
        the end of the genome.
        """
        self.assertTrue(
            GenomeRanges("join{[20:30](+), [31:40](+), [0:10](+)}").circular(40)
        )

    def testThreeRangesThatAreCircular2(self):
        """
        The circular method must return True when given three ranges that span
        the end of the genome when one of the ranges is the last in the passed
        BioPython GenBank string and the second range (starting at zero) is the
        first.
        """
        self.assertTrue(
            GenomeRanges("join{[0:10](+), [40:60](+), [80:90](+)}").circular(90)
        )


class TestStartInGenome(TestCase):
    """
    Test the GenomeRanges class startInGenome method.
    """

    def testStartOutsideGenomeRange(self):
        """
        If the start location in the DIAMOND match is greater than the length
        of the genome, a ValueError must be raised.
        """
        # The starting offset in the genome is 27 because DIAMOND returns
        # 1-based offsets into the protein subject. So in Python terms the
        # start offset is amino acid 9, or nucleotide 27.
        ranges = GenomeRanges("[10:20](+)")
        error = (
            r"^Starting nucleotide offset 27 not found in protein "
            r"nucleotide ranges \(10, 20\)\.$"
        )
        self.assertRaisesRegex(ValueError, error, ranges.startInGenome, {"sstart": 10})

    def testStartOutsideGenomeRangeTwoRanges(self):
        """
        If the start location in the DIAMOND match is greater than the length
        of the genome when two ranges are given, a ValueError must be raised.
        """
        # The starting offset in the genome is 27 because DIAMOND returns
        # 1-based offsets into the protein subject. So in Python terms the
        # start offset is amino acid 9, or nucleotide 27.
        ranges = GenomeRanges("join{[5:10](+), [12:18](+)}")
        error = (
            r"^Starting nucleotide offset 27 not found in protein "
            r"nucleotide ranges \(5, 10\), \(12, 18\)\.$"
        )
        self.assertRaisesRegex(ValueError, error, ranges.startInGenome, {"sstart": 10})

    def testMatchAtStartOfFirstRange(self):
        """
        If the start location in the DIAMOND match is at the very beginning of
        the first range, the correct offset must be returned.
        """
        # The offset is 5 because the match starts 0 nucleotides into the
        # protein (0 = (1 - 1 ) * 3) and the protein starts at position 5 in
        # the genome. So the match begins at nucleotide 0 + 5 = 5.
        ranges = GenomeRanges("[5:35](+)")
        self.assertEqual(5, ranges.startInGenome({"sstart": 1}))

    def testMatchAtEndOfFirstRange(self):
        """
        If the start location in the DIAMOND match is at the very end of
        the first range, the correct offset must be returned.
        """
        # The offset is 20 because the match starts 15 nucleotides into the
        # protein (15 = (6 - 1 ) * 3) and the protein starts at position 5 in
        # the genome. So the match begins at nucleotide 15 + 5 = 20.
        ranges = GenomeRanges("[5:21](+)")
        self.assertEqual(20, ranges.startInGenome({"sstart": 6}))

    def testInFirstRange(self):
        """
        If the start location in the DIAMOND match is in the first range,
        the correct offset must be returned.
        """
        # The offset is 32 because the match starts 27 nucleotides into the
        # protein (27 = (10 - 1 ) * 3) and the protein starts at position 5 in
        # the genome. So the match begins at nucleotide 27 + 5 = 32.
        ranges = GenomeRanges("[5:35](+)")
        self.assertEqual(32, ranges.startInGenome({"sstart": 10}))

    def testMatchAtStartOfSecondRange(self):
        """
        If the start location in the DIAMOND match is at the very beginning
        of the second range, the correct offset must be returned.
        """
        # The offset is 50 because the match starts 30 nucleotides into the
        # protein (45 = (16 - 1 ) * 3) and the protein has a range of 30
        # nucleotides (35 - 5 = 30) and then a range of 35 nucleotides (85 -
        # 50 = 35). So the match begins 15 (45 - 30 = 15) nucleotides into
        # the second range (which starts at 50), and 15 + 50 = 65.
        ranges = GenomeRanges("join{[5:35](+), [50:85](+)}")
        self.assertEqual(50, ranges.startInGenome({"sstart": 11}))

    def testInSecondRange(self):
        """
        If the start location in the DIAMOND match is in the second range,
        the correct offset must be returned.
        """
        # The offset is 65 because the match starts 45 nucleotides into the
        # protein (45 = (16 - 1 ) * 3) and the protein has a range of 30
        # nucleotides (35 - 5 = 30) and then a range of 35 nucleotides (85 -
        # 50 = 35). So the match begins 15 (45 - 30 = 15) nucleotides into
        # the second range (which starts at 50), and 15 + 50 = 65.
        ranges = GenomeRanges("join{[5:35](+), [50:85](+)}")
        self.assertEqual(65, ranges.startInGenome({"sstart": 16}))

    def testInThirdRange(self):
        """
        If the start location in the DIAMOND match is in the third range,
        the correct offset must be returned.
        """
        # The offset is 2900 because the match starts 1200 nucleotides into
        # the protein (1200 = (401 - 1 ) * 3) and the protein has a range of
        # 100 nucleotides (100 - 0 = 100), then a range of 200 nucleotides
        # (600 - 400 = 200), then a range of 1000 nucleotides. So the match
        # begins 900 (1200 - 300 = 900) nucleotides into the third range
        # (which starts at 2000), and 2000 + 900 = 2900.
        ranges = GenomeRanges("join{[0:100](+), [400:600](+), [2000:3000](+)}")
        self.assertEqual(2900, ranges.startInGenome({"sstart": 401}))


class TestOrientations(TestCase):
    """
    Test the GenomeRanges class orientations method.
    """

    def testAllComplement(self):
        """
        If all ranges are on the complement strand, the orientations method
        must return just True.
        """
        ranges = GenomeRanges("join{[0:100](-), [400:600](-)}")
        self.assertEqual({False}, ranges.orientations())

    def testNoComplement(self):
        """
        If no ranges are on the complement strand, the orientations method
        must return just False.
        """
        ranges = GenomeRanges("join{[0:100](+), [400:600](+)}")
        self.assertEqual({True}, ranges.orientations())

    def testMixed(self):
        """
        If the ranges are on the both the regular and the complement strand,
        the orientations method must return the set {True, False}.
        """
        ranges = GenomeRanges("join{[0:100](+), [400:600](-)}")
        self.assertEqual({True, False}, ranges.orientations())


class TestDistinctRangeCount(TestCase):
    """
    Test the GenomeRanges class distinctRangeCount method.
    """

    def testOneRange(self):
        """
        If there is only one range, 1 must be returned.
        """
        ranges = GenomeRanges("[0:100](+)")
        self.assertEqual(1, ranges.distinctRangeCount(100))

    def testTwoRange(self):
        """
        If there are 2 ranges, 2 must be returned.
        """
        ranges = GenomeRanges("join{[0:100](+), [400:600](+)}")
        self.assertEqual(2, ranges.distinctRangeCount(100))

    def testThreeRange(self):
        """
        If there are 3 ranges, 3 must be returned.
        """
        ranges = GenomeRanges("join{[0:100](+), [400:600](+), [2000:3000](+)}")
        self.assertEqual(3, ranges.distinctRangeCount(100))

    def testThreeRangesCircular(self):
        """
        If there are 3 ranges but the genome is circular, 2 must be returned.
        """
        ranges = GenomeRanges("join{[0:100](+), [400:600](+), [2000:3000](+)}")
        self.assertEqual(2, ranges.distinctRangeCount(3000))



#!/usr/bin/env python

import sys
import os
from time import time
from itertools import chain

from dark.fasta import SqliteIndex


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Create an sqlite3 database from FASTA sequences."
    )

    parser.add_argument(
        "--out",
        required=True,
        help=(
            "The output file. This file must not exist (use --force to " "overwrite)."
        ),
    )

    parser.add_argument(
        "--force",
        default=False,
        action="store_true",
        help="If True and the output file already exists, overwrite it.",
    )

    parser.add_argument(
        "--quiet",
        default=False,
        action="store_true",
        help="If True do not print indexing progress.",
    )

    parser.add_argument(
        "--fasta",
        metavar="FASTA-file",
        nargs="+",
        action="append",
        required=True,
        help=(
            "the FASTA file(s) to make the database from. These may be "
            "uncompressed, or compressed with bgzip (from samtools), with "
            "a .gz suffix."
        ),
    )

    args = parser.parse_args()

    if os.path.exists(args.out):
        if args.force:
            os.unlink(args.out)
        else:
            print(
                "Output file '%s' already exists. Use --force to overwrite." % args.out,
                file=sys.stderr,
            )
            sys.exit(1)

    index = SqliteIndex(args.out)

    # Flatten the lists of lists that we get from using both nargs='+' and
    # action='append'. We use both because it allows people to use (e.g.)
    # --fasta on the command line either via "--fasta file1 --fasta file2"
    # or "--fasta file1 file2", or a combination of these. That way it's
    # not necessary to remember which way you're supposed to use it and you
    # also can't be hit by the subtle problem encountered in
    # https://github.com/acorg/dark-matter/issues/453
    fastaFiles = list(chain.from_iterable(args.fasta))

    verbose = not args.quiet

    for filename in fastaFiles:
        if verbose:
            print("Indexing '%s' ... " % filename, end="", file=sys.stderr)
            start = time()

        count = index.addFile(filename)

        if verbose:
            elapsed = time() - start
            print(
                "indexed %d sequence%s in %.2f seconds."
                % (count, "" if count == 1 else "s", elapsed),
                file=sys.stderr,
            )



import sys
from time import time, ctime
from typing import Optional, Union, TextIO
from subprocess import PIPE, CompletedProcess, CalledProcessError, run
from io import StringIO


class Executor:
    """
    Log and execute shell commands.

    @param dryRun: If C{True}, do not execute commands, just log them. This sets the
        default and can be overidden for a specific command by passing C{dryRun} to
        the C{execute} method.
    @stdout: If not C{None}, must be an open file descriptor. Both the commands that
        are executed and their standard output will be written to this descriptor.
    @stderr: If not C{None}, must be an open file descriptor. The standard error output
        of commands will be written to this descriptor.
    """

    def __init__(
        self,
        dryRun: bool = False,
        stdout: Optional[Union[TextIO, StringIO]] = None,
        stderr: Optional[Union[TextIO, StringIO]] = None,
    ) -> None:
        self.dryRun = dryRun
        self.stdout = stdout
        self.stderr = stderr
        self.log = [f"# Executor created at {ctime(time())}. Dry run = {dryRun}."]

    def execute(
        self,
        command: Union[str, list[str]],
        dryRun: Optional[bool] = None,
        useStderr: bool = True,
        stdout: Optional[Union[bool, TextIO, StringIO]] = False,
        stderr: Optional[Union[bool, TextIO, StringIO]] = False,
        **kwargs,
    ) -> Optional[CompletedProcess]:
        """
        Execute (or simulate) a command. Add to our log.

        @param command: Either a C{str} command (which will be passed to the
            shell) or a C{list} of command arguments (including the executable
            name), in which case the shell is not used.
        @param dryRun: If C{True}, do not execute commands, just log them.
            If C{False}, execute the commands. If not given or C{None}, use
            the default setting (in C{self._dryRun}).
        @param useStderr: If C{True} and the command results in an exception, print
            a summary of the command standard output and standard error to sys.stderr.
            If a function is passed, the exception is passed to the function and the
            summary is printed to sys.stderr if the function returns C{True}.
        @stdout: If not C{None}, must be an open file descriptor. Both the
            commands that are executed and their standard output will be written
            to this descriptor. If C{None}, output will not be written to any
            descriptor (but remains available on the result object, as usual). If no
            value is passed, the value originally passed to __init__ will be used.
        @stderr: If not C{None}, must be an open file descriptor. The standard
            error output of commands will be written to this descriptor.  If C{None},
            stderr output will not be written to any descriptor (but remains available
            on the result object, as usual). If no value is passed, the value
            originally passed to __init__ will be used.
        @param kwargs: Keyword arguments that will be passed to subprocess.run. Note
            that keyword arguments are not currently logged (the logging is slightly
            problematic since a keyword argument might be an environment dictionary).
        @raise CalledProcessError: If the command results in an error.
        @return: A C{CompletedProcess} instance. This has attributes such as
            C{returncode}, C{stdout}, and C{stderr}. See pydoc subprocess.
            If C{dryRun} is C{True}, C{None} is returned.
        """
        stdout = self.stdout if stdout is False else stdout
        stderr = self.stderr if stderr is False else stderr

        if isinstance(command, str):
            # Can't have newlines in a command given to the shell.
            strCommand = command = command.replace("\n", " ").strip()
            shell = True
        else:
            strCommand = " ".join(command)
            shell = False

        if isinstance(stdout, (StringIO, TextIO)):
            print("$ " + strCommand, file=stdout)

        dryRun = self.dryRun if dryRun is None else dryRun

        if dryRun:
            self.log.append("$ " + strCommand)
            return None

        start = time()
        self.log.extend(
            [
                "# Start command (shell=%s) at %s" % (shell, ctime(start)),
                "$ " + strCommand,
            ]
        )

        try:
            result = run(
                command,
                check=True,
                stdout=PIPE,
                stderr=PIPE,
                shell=shell,
                universal_newlines=True,
                **kwargs,
            )
        except CalledProcessError as e:
            if callable(useStderr):
                useStderr = useStderr(e)
            if useStderr:
                print("CalledProcessError:", e, file=sys.stderr)
                print("STDOUT:\n%s" % e.stdout, file=sys.stderr)
                print("STDERR:\n%s" % e.stderr, file=sys.stderr)
            raise

        stop = time()
        elapsed = stop - start
        self.log.extend(
            [
                "# Stop command at %s" % ctime(stop),
                "# Elapsed = %f seconds" % elapsed,
            ]
        )

        if result.stdout and isinstance(stdout, (StringIO, TextIO)):
            print(result.stdout, end="", file=stdout)

        if result.stderr and isinstance(stderr, (StringIO, TextIO)):
            print(result.stderr, end="", file=stderr)

        return result
